
import copy

from ..symrefs import symref
from ..formulas import CompoundFormula, QuantifiedFormula, Atom, Formula
from ..terms import Term, CompoundTerm
from .errors import SubstitutionError


class TermSubstitution:
    """ This Visitor applies a substitution operation on a given formula, replacing terms in the dictionary
        'subst' by the associated term. """

    def __init__(self, lang, subst):
        self.lang = lang
        self.subst = subst

    def visit(self, phi):
        if isinstance(phi, CompoundFormula):
            _ = [self.visit(f) for f in phi.subformulas]

        elif isinstance(phi, QuantifiedFormula):
            if any(symref(x) in self.subst for x in phi.variables):
                raise SubstitutionError(phi, self.subst, 'Attempted to substitute variable bound by quantifier')
            self.visit(phi.formula)

        elif isinstance(phi, (Atom, CompoundTerm)):
            new_subterms = list(phi.subterms)
            for k, t in enumerate(new_subterms):
                rep = self.subst.get(symref(t), None)
                if rep is None:
                    self.visit(t)
                else:
                    new_subterms[k] = rep
            phi.subterms = tuple(new_subterms)


def term_substitution(language, phi, substitution, inplace=False):
    """ Return the result of applying the given substitution to the given formula or term of the language.
    If `inplace` is true, the given formula is the one modified.
    """
    assert isinstance(phi, (Formula, Term))
    phi = phi if inplace else copy.deepcopy(phi)
    op = TermSubstitution(language, substitution)
    op.visit(phi)
    return phi


def create_substitution(symbols, values):
    return {symref(symbols[k]): v for k, v in enumerate(values)}




from typing import Tuple

from .util import termlists_are_equal, termlist_hash
from .sorts import Sort, parent
from .. import errors as err
from .builtins import BuiltinPredicateSymbol, BuiltinFunctionSymbol


class Term:
    """ A first-order logical term.

        Important note: Term objects overload the equality operator `__eq__` to provide syntactic sugar and allow the
        construction of FOL atoms such as "loc(b1)==table". A side-effect of this is that Term objects cannot be
        inserted as such into associative containers such as dictionaries or sets. In order to use those, you will
        need to wrap the Term object with a call to `symref`, as e.g. in:
        >>> from tarski.fstrips import language
        >>> from tarski.syntax import symref
        >>> lang = language("test")
        >>> counter = dict()
        >>> c = lang.constant('c', 'object')
        >>> counter[c] = 2  # ERROR: will not work correctly
        >>> counter[symref(c)] = 2  # This is the correct way of doing it

        To prevent usage in such containers, Term objects are not hashable. If you need to hash them for other purposes,
        use the `t.hash()` method.
    """

    @property
    def language(self):
        raise NotImplementedError()  # Let the subclasses implement this

    @property
    def sort(self):
        raise NotImplementedError()  # Let the subclasses implement this

    # required to support scoped declaration of variables
    # with bw.var('x',block) as x, bw.var('y',block) as y do :
    #   ... declarations using x and y
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return exc_type is not None

    def __lshift__(self, rhs):
        return self.language.dispatch_operator('<<', Term, Term, self, rhs)

    def __rshift__(self, rhs):
        return self.language.dispatch_operator('>>', Term, Term, self, rhs)

    def __eq__(self, rhs):
        return self.language.dispatch_operator(BuiltinPredicateSymbol.EQ, Term, Term, self, rhs)

    def __ne__(self, rhs):
        return self.language.dispatch_operator(BuiltinPredicateSymbol.NE, Term, Term, self, rhs)

    def __lt__(self, rhs):
        return self.language.dispatch_operator(BuiltinPredicateSymbol.LT, Term, Term, self, rhs)

    def __gt__(self, rhs):
        return self.language.dispatch_operator(BuiltinPredicateSymbol.GT, Term, Term, self, rhs)

    def __le__(self, rhs):
        return self.language.dispatch_operator(BuiltinPredicateSymbol.LE, Term, Term, self, rhs)

    def __ge__(self, rhs):
        return self.language.dispatch_operator(BuiltinPredicateSymbol.GE, Term, Term, self, rhs)

    def __add__(self, rhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.ADD, Term, Term, self, rhs)

    def __sub__(self, rhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.SUB, Term, Term, self, rhs)

    def __mul__(self, rhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.MUL, Term, Term, self, rhs)

    def __matmul__(self, rhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.MATMUL, Term, Term, self, rhs)

    def __truediv__(self, rhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.DIV, Term, Term, self, rhs)

    def __radd__(self, lhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.ADD, Term, Term, lhs, self)

    def __rsub__(self, lhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.SUB, Term, Term, lhs, self)

    def __rmul__(self, lhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.MUL, Term, Term, lhs, self)

    def __rtruediv__(self, lhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.DIV, Term, Term, lhs, self)

    def __pow__(self, rhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.POW, Term, Term, self, rhs)

    def __mod__(self, rhs):
        return self.language.dispatch_operator(BuiltinFunctionSymbol.MOD, Term, Term, self, rhs)

    def __floordiv__(self, rhs):
        return self.language.dispatch_operator('//', Term, Term, self, rhs)

    def __divmod__(self, rhs):
        return self.language.dispatch_operator('divmod', Term, Term, self, rhs)

    def __and__(self, rhs):
        return self.language.dispatch_operator('&', Term, Term, self, rhs)

    def __xor__(self, rhs):
        return self.language.dispatch_operator('^', Term, Term, self, rhs)

    def __or__(self, rhs):
        return self.language.dispatch_operator('|', Term, Term, self, rhs)

    def is_syntactically_equal(self, other):
        """ Return true if this term and other are strictly syntactically equivalent.
        This would tipically be in the magic method `__eq__`, but we use `__eq__` for a different purpose,
        namely, to be able to use Python expressions such as "loc(b1)==table" in order to construct a FOL atom. """
        raise NotImplementedError()  # To be subclassed

    # def __hash__(self):
    #     Raise an informative exception to prevent wrong usage of terms in associative containers
        # raise err.WrongTermUsageError()
    # @see https://docs.python.org/3/reference/datamodel.html#object.__hash__
    __hash__ = None

    def hash(self):
        """ Return a hash of the current object. Meant to be used by TermReference objects to wrap Terms appropriately
        for use within associative containers. """
        raise NotImplementedError()  # To be subclassed


class Variable(Term):

    def __init__(self, symbol: str, sort: Sort):
        self.symbol = symbol
        self._sort = sort
        self.sort.language.language_components_frozen = True
        # TODO VALIDATE

    @property
    def language(self):
        return self.sort.language

    @property
    def sort(self):
        return self._sort

    def __str__(self):
        return '{}/{}'.format(self.symbol, self.sort.name)

    __repr__ = __str__

    def hash(self):
        return hash((self.symbol, self.sort.name))

    def is_syntactically_equal(self, other):
        return self.__class__ is other.__class__ and self.symbol == other.symbol and self.sort.name == other.sort.name


class CompoundTerm(Term):
    """ A first order compound term such as "f(c, d)", where f is a function symbol of arity 2, and c and d are both
    constant symbols. More generally, a compound term is a function symbol of a certain arity n paired with an n-tuple
    of terms, sorts matching.
    """

    def __init__(self, symbol, subterms: Tuple[Term]):

        self.symbol = symbol

        if len(subterms) != self.symbol.arity:
            raise err.ArityMismatch(symbol, subterms)
        argument_sorts = list(self.symbol.sort)[:-1]
        processed_st = []
        for k, s in enumerate(argument_sorts):
            # @TODO Implement upcasting for non built-in compound terms
            try:
                if subterms[k].sort.name != s.name and not self.symbol.language.is_subtype(subterms[k].sort, s):
                    raise err.SortMismatch(self.symbol, subterms[k].sort, s)
                processed_st.append(subterms[k])
            except AttributeError:
                s_k = s.cast(subterms[k])
                if s_k is None:
                    raise err.SortMismatch(self.symbol, subterms[k], s)
                processed_st.append(s_k)
        self.subterms = tuple(processed_st)
        self.symbol.language.language_components_frozen = True

    @property
    def language(self):
        return self.symbol.language

    @property
    def sort(self):
        return self.symbol.codomain

    def __str__(self):
        return '{}({})'.format(self.symbol.symbol, ', '.join([str(t) for t in self.subterms]))

    __repr__ = __str__

    def hash(self):
        return hash((self.symbol.symbol, termlist_hash(self.subterms)))

    def is_syntactically_equal(self, other):
        return self.__class__ is other.__class__ and \
               self.symbol == other.symbol and \
               termlists_are_equal(self.subterms, other.subterms)


class AggregateCompoundTerm(Term):
    """
        Combinations of a functional symbol, a set of bound variables and a sequence of terms.
        Examples of such terms include summations, products and other operations over sequences.

        Argument symbol needs to be an instance of Function or have Function-like interface.
    """

    def __init__(self, operator, bound_vars, subterm: Term):
        self.symbol = operator
        self.bound_vars = bound_vars
        self.subterm = subterm  # TODO: type checking?

    @property
    def language(self):
        return self.subterm.language

    @property
    def sort(self):
        return self.subterm.sort

    def __str__(self):
        return '{}_{{{}}}({})'.format(self.symbol, ','.join([str(x) for x in self.bound_vars]), str(self.subterm))

    __repr__ = __str__

    def hash(self):
        raise NotImplementedError()

    def is_syntactically_equal(self, other):
        return self.__class__ is other.__class__ and \
               self.symbol == other.symbol and \
               termlists_are_equal(self.bound_vars, other.bound_vars) and \
               self.subterm.is_syntactically_equal(other.subterm)


class IfThenElse(Term):
    """
        Combination of a formula C and two terms, t1 and t2.

        This term evaluates to t1 if C is true, and t2 otherwise. t1 and t2
        are restricted to have the same codomain.
    """

    def __init__(self, condition, subterms: Tuple[Term]):
        self.symbol = subterms[0].language.get('ite')
        self.condition = condition
        if len(subterms) != 2:
            raise err.ArityMismatch(self.symbol, subterms, msg='IfThenElse: needs two sub terms!')

        # Our implementation of ite requires both branches to have equal sort
        if subterms[0].sort != subterms[1].sort:
            if parent(subterms[0].sort) == subterms[1].sort:
                self._sort = subterms[1].sort
            elif parent(subterms[1].sort) == subterms[0].sort:
                self._sort = subterms[0].sort
            else:
                raise err.SyntacticError(
                    msg='IfThenElse: both subterms need to be of the same sort! lhs: "{}"({}), rhs: "{}"({})'.format(
                        subterms[0], subterms[0].sort, subterms[1], subterms[1].sort))
        else:
            self._sort = subterms[0].sort

        self.subterms = tuple(subterms)

        self.symbol.language.language_components_frozen = True

    @property
    def language(self):
        return self.symbol.language

    @property
    def sort(self):
        return self._sort

    def __str__(self):
        return '{}({})'.format(self.symbol, ', '.join([str(self.condition)] + [str(t) for t in self.subterms]))

    __repr__ = __str__

    def hash(self):
        return hash(('ite', self.condition, termlist_hash(self.subterms)))

    def is_syntactically_equal(self, other):
        return self.__class__ is other.__class__ and \
               self.symbol == other.symbol and \
               self.condition.is_syntactically_equal(other.condition) and \
               termlists_are_equal(self.subterms, other.subterms)


def ite(c, t1: Term, t2: Term):
    return IfThenElse(c, (t1, t2))


class Constant(Term):
    def __init__(self, symbol, sort: Sort):
        self.symbol = symbol
        self._sort = sort
        # symbol validation
        if not self._sort.builtin:
            # construction of constants extends the domain of sorts
            self._sort.extend(self)
        else:
            self.symbol = self._sort.cast(self.symbol)
        self.sort.language.language_components_frozen = True

    @property
    def language(self):
        return self.sort.language

    @property
    def sort(self):
        return self._sort

    def __str__(self):
        return '{}'.format(self.symbol)

    def __repr__(self):
        return '{} ({})'.format(self.symbol, self.sort.name)

    def hash(self):
        return hash((self.symbol, self.sort))

    def is_syntactically_equal(self, other):
        return self.__class__ is other.__class__ and self.symbol == other.symbol and self.sort == other.sort



"""
 Structural CSP decomposition methods
"""
import itertools
from collections import defaultdict

from ..errors import TarskiError
from ..syntax import CompoundFormula, Atom, Connective, Variable, Constant


class WrongFormalismError(TarskiError):
    pass


def compute_schema_constraint_hypergraph(action):
    """ Compute the constraint hypergraph corresponding to the precondition of the given action schema.
    This is the hypergraph with one node v for each action parameter, and one hyperedge {v1, ..., vn} for each atom
    p(v1, ..., vn) in the conjunctive precondition of the action schema """
    hyperedges = set()
    _collect_hyperedges(action.precondition, hyperedges)
    return hyperedges


def compute_schema_primal_graph(action):
    """ Compute the primal graph corresponding to the precondition of the given action schema.
    This is the graph with one node v for each action parameter, and one edge {v1, v2} for every pair of variables
    v1, v2 that appear together in some atom p(..., v1, ..., v2, ...) in the conjunctive precondition of the schema. """
    return compute_primal_graph_from_hypergraph(compute_schema_constraint_hypergraph(action))


def compute_primal_graph_from_hypergraph(hypergraph):
    """ Compute the primal graph corresponding to the given hypergraph. """
    edges = set()
    for hyperedge in hypergraph:
        edges.update(itertools.combinations(hyperedge, 2))
    return edges


def _remove_ear_if_exists(edges, node_counts):
    """ A helper to find a so-called ear of the hypergraph """
    for e1, e2 in itertools.product(edges, edges):
        diff = e1-e2
        if diff and all(node_counts[node] == 1 for node in diff):
            # e1 is an ear and can be removed. We update node count as well
            for node in e1:
                node_counts[node] -= 1
            edges.remove(e1)
            return True
    return False


def check_hypergraph_acyclicity(hypergraph):
    """ Check whether the given hypergraph is acyclic by applying the GYO reduction as described in
        Jeffrey D. Ullman, Principles of Database and Knowledge-Base Systems, Vol. II
    """
    nodes = set(itertools.chain.from_iterable(hypergraph))
    edges = set(frozenset(x) for x in hypergraph)  # simply convert the tuple into frozensets
    if len(edges) <= 1 or len(nodes) <= 1:
        return True

    # Compute a mapping storing in how many hyperedges each node appears
    node_counts = defaultdict(int)
    for edge in edges:
        for n in edge:
            node_counts[n] += 1

    while _remove_ear_if_exists(edges, node_counts):
        pass

    return len(edges) <= 1


def _collect_hyperedges(phi, edges):
    if isinstance(phi, Atom):
        assert all(isinstance(x, (Variable, Constant)) for x in phi.subterms)
        edge = set()
        for x in phi.subterms:
            if isinstance(x, Variable):
                edge.add(x.symbol[1:])  # Get rid of the "?" in the variable name
        if edge:  # Discard nullary and unary atoms
            edges.add(tuple(sorted(edge)))  # Sort to prevent counting both (x,y) and (y,x) as different hyperedges
    elif isinstance(phi, CompoundFormula):
        if phi.connective == Connective.And:
            _ = [_collect_hyperedges(sub, edges) for sub in phi.subformulas]
        elif phi.connective == Connective.Not:
            _collect_hyperedges(phi.subformulas[0], edges)
        else:
            raise WrongFormalismError("Can only process conjunctive queries")
    else:
        raise WrongFormalismError("Can only process conjunctive queries")



"""
    CNF Transformation
"""
from ..formulas import CompoundFormula, Connective, QuantifiedFormula
from ..transform import NNFTransformation

from .errors import TransformationError


class CNFTransformation:
    """
        This class rewrites an input quantifier free formula phi into an equivalent formula
        in CNF

    """

    def __init__(self, lang, phi, do_copy=True):
        """ Distribute w.r.t. disjunction, see Huth & Ryan, pp. 60-62 """
        self.L = lang
        self.blueprint = NNFTransformation.rewrite(phi, do_copy).nnf
        self.cnf = None
        self.clauses = []
        self.current_clause = []

    def distribute(self, n1, n2):
        if isinstance(n1, CompoundFormula) and n1.connective == Connective.And:
            lhs = self.distribute(n1.subformulas[0], n2)
            rhs = self.distribute(n1.subformulas[1], n2)
            return CompoundFormula(Connective.And, (lhs, rhs))
        elif isinstance(n2, CompoundFormula) and n2.connective == Connective.And:
            lhs = self.distribute(n1, n2.subformulas[0])
            rhs = self.distribute(n1, n2.subformulas[1])
            return CompoundFormula(Connective.And, (lhs, rhs))
        else:
            return CompoundFormula(Connective.Or, (n1, n2))

    def _convert(self, phi):

        if isinstance(phi, QuantifiedFormula):
            raise TransformationError("cnf transformation", phi, "Formula is not quantifier free!")

        if isinstance(phi, CompoundFormula):
            if phi.connective == Connective.Not:
                return phi  # already CNF
            elif phi.connective == Connective.And:
                lhs, rhs = self._convert(phi.subformulas[0]), self._convert(phi.subformulas[1])
                return CompoundFormula(Connective.And, (lhs, rhs))
            else:
                assert phi.connective == Connective.Or
                lhs, rhs = self._convert(phi.subformulas[0]), self._convert(phi.subformulas[1])
                return self.distribute(lhs, rhs)

        return phi

    def collect_clauses(self, phi):
        if isinstance(phi, QuantifiedFormula):
            raise TransformationError("cnf transformation", phi, "Formula is not quantifier free!")

        if isinstance(phi, CompoundFormula):
            if phi.connective == Connective.Not:
                self.current_clause.append(phi)
                return None
            elif phi.connective == Connective.And:
                self.current_clause = []
                self.collect_clauses(phi.subformulas[0])
                self.clauses.append(self.current_clause)
                self.current_clause = []
                self.collect_clauses(phi.subformulas[1])
                self.clauses.append(self.current_clause)
                return None
            else:
                self.collect_clauses(phi.subformulas[0])
                self.collect_clauses(phi.subformulas[1])
                return None
        self.current_clause.append(phi)
        return phi

    def convert(self):
        self.cnf = self._convert(self.blueprint)
        self.collect_clauses(self.cnf)

    @staticmethod
    def rewrite(lang, phi, do_copy=True):
        trans = CNFTransformation(lang, phi, do_copy)
        trans.convert()
        return trans



import numpy as np
import scipy.special as sci

impl = {
    "min": lambda x, y: np.min((x, y)),
    "max": lambda x, y: np.max((x, y)),
    "pow": np.power,
    "abs": np.abs,
    "sin": np.sin,
    "cos": np.cos,
    "tan": np.tan,
    "asin": np.arcsin,
    "atan": np.arctan,
    "exp": np.exp,
    "log": np.log,
    "erf": sci.erf,
    "erfc": sci.erfc,
    "sgn": np.sign,
    "sqrt": np.sqrt,
    "normal": np.random.normal,
    "gamma": np.random.gamma,
}



import itertools
import copy
import numpy as np

from ...syntax import Term, AggregateCompoundTerm, CompoundTerm, Constant, Variable, IfThenElse, create_substitution, \
    term_substitution
from ...syntax.algebra import Matrix
from ... import errors as err
from ... grounding.naive import instantiation
from ..builtins import BuiltinFunctionSymbol, get_arithmetic_binary_functions


def sumterm(*args):
    variables = args[:-1]
    expr = args[-1]
    if len(variables) < 1:
        raise err.SyntacticError(msg='sumterm(x0,x1,...,xn,expr) requires at least one\
        bound variable, arguments: {}'.format(args))
    for x in variables:
        if not isinstance(x, Variable):
            raise err.SyntacticError(msg='sum(x0,...,xn,expr) require each\
            argument xi to be an instance of Variable')
    if not isinstance(expr, Term):
        raise err.SyntacticError(msg='sum(x0,x1,...,xn,expr) requires last \
        argument "expr" to be an instance of Term, got "{}"'.format(expr))
    return AggregateCompoundTerm(BuiltinFunctionSymbol.ADD, variables, expr)


def prodterm(*args):
    variables = args[:-1]
    expr = args[-1]
    if len(variables) < 1:
        raise err.SyntacticError(msg='sumterm(x0,x1,...,xn,expr) requires at least one\
        bound variable, arguments: {}'.format(args))
    for x in variables:
        if not isinstance(x, Variable):
            raise err.SyntacticError(msg='sum(x0,...,xn,expr) require each\
            argument xi to be an instance of Variable')
    if not isinstance(expr, Term):
        raise err.SyntacticError(msg='sum(x0,x1,...,xn,expr) requires last \
        argument "expr" to be an instance of Term, got "{}"'.format(expr))
    return AggregateCompoundTerm(BuiltinFunctionSymbol.MUL, variables, expr)


def summation(*args):
    """
        Summation of a (nested) sequence of expressions defined over one or more
        variables.
    """
    expr = args[-1]
    if not isinstance(expr, Term):
        raise err.SyntacticError(msg='sum(x0,x1,...,xn,expr) requires last \
        argument "expr" to be an instance of Term, got "{}"'.format(expr))
    variables = []
    for x in args[:-1]:
        if not isinstance(x, Variable):
            raise err.SyntacticError(msg='sum(x0,...,xn,expr) require each\
            argument xi to be an instance of Variable')
        variables.append(x)

    L = expr.language
    _, syms, substs = instantiation.enumerate_groundings(L, list(variables))
    processed_expr = []
    for values in itertools.product(*substs):
        subst = create_substitution(syms, values)
        processed_expr.append(term_substitution(L, expr, subst))

    lhs = processed_expr[0]
    for k in range(1, len(processed_expr)):
        lhs = L.dispatch_operator(BuiltinFunctionSymbol.ADD, Term, Term, lhs, processed_expr[k])

    return lhs


def product(*args):
    """
        Product of a (nested) sequence of expressions defined over one or
        more variables.
    """
    expr = args[-1]
    if not isinstance(expr, Term):
        raise err.SyntacticError(msg='prod(x0,x1,...,xn,expr) requires last \
        argument "expr" to be an instance of Term')
    variables = []
    for x in args[:-1]:
        if not isinstance(x, Variable):
            raise err.SyntacticError(msg='prod(x0,...,xn,expr) require each\
            argument xi to be an instance of Variable')
        variables.append(x)

    L = expr.language
    _, syms, substs = instantiation.enumerate_groundings(L, list(variables))
    processed_expr = []
    for values in itertools.product(*substs):
        subst = create_substitution(syms, values)
        processed_expr.append(term_substitution(L, expr, subst))

    lhs = processed_expr[0]
    for k in range(1, len(processed_expr)):
        lhs = L.dispatch_operator(BuiltinFunctionSymbol.MUL, Term, Term, lhs, processed_expr[k])

    return lhs


def pow(x, y):
    pow_func = x.language.get_function(BuiltinFunctionSymbol.POW)
    return pow_func(x, y)


def sqrt(x):
    sqrt_func = x.language.get_function(BuiltinFunctionSymbol.SQRT)
    return sqrt_func(x)


def transpose(m: Term):
    if isinstance(m, Matrix):
        m_t = copy.copy(m)
        m_t.matrix = m_t.matrix.T
        return m_t
    elif isinstance(m, CompoundTerm):
        if m.symbol.symbol in get_arithmetic_binary_functions():
            # handle as an operator
            if m.symbol.symbol == BuiltinFunctionSymbol.ADD:
                m.subterms = (transpose(m.subterms[0]), transpose(m.subterms[1]))
                return m
            if m.symbol.symbol == BuiltinFunctionSymbol.SUB:
                m.subterms = (transpose(m.subterms[0]), transpose(m.subterms[1]))
                return m
            elif m.symbol.symbol == BuiltinFunctionSymbol.MUL:
                m.subterms = (transpose(m.subterms[1]), transpose(m.subterms[0]))
                return m
        raise err.SyntacticError("transpose() only applicable on scalars (constants, variables), "
                                 "matrices and basic arithmetic operations.")
    elif isinstance(m, Constant):
        return Matrix([m], m.sort)
    elif isinstance(m, Variable):
        return Matrix([m], m.sort)
    raise err.SyntacticError(
        "transpose(): can only be applied on scalars (constants, variables), matrices and basic arithmetic operations.")


def zero(sort):
    assert sort.name in ('Real', 'Integer', 'Natural')
    return sort.language.constant(0, sort)


def one(sort):
    assert sort.name in ('Real', 'Integer', 'Natural')
    return sort.language.constant(1, sort)


def simplify(expr: Term):
    if isinstance(expr, Constant):
        return expr
    elif isinstance(expr, Variable):
        return expr
    elif isinstance(expr, CompoundTerm):
        if not expr.symbol.builtin:
            return expr
        if expr.symbol.symbol == BuiltinFunctionSymbol.ADD:
            simp_st = (simplify(expr.subterms[0]), simplify(expr.subterms[1]))
            if simp_st[0].is_syntactically_equal(zero(expr.sort)):
                return simp_st[1]
            if simp_st[1].is_syntactically_equal(zero(expr.sort)):
                return simp_st[0]
            expr.subterms = simp_st
            return expr
        if expr.symbol.symbol == BuiltinFunctionSymbol.SUB:
            simp_st = (simplify(expr.subterms[0]), simplify(expr.subterms[1]))
            if simp_st[0].is_syntactically_equal(zero(expr.sort)):
                return -1 * simp_st[1]
            if simp_st[1].is_syntactically_equal(zero(expr.sort)):
                return simp_st[0]
            expr.subterms = simp_st
            return expr
        if expr.symbol.symbol == BuiltinFunctionSymbol.MUL:
            simp_st = (simplify(expr.subterms[0]), simplify(expr.subterms[1]))
            if simp_st[0].is_syntactically_equal(zero(expr.sort)):
                return zero(expr.sort)
            if simp_st[1].is_syntactically_equal(zero(expr.sort)):
                return zero(expr.sort)
            if simp_st[0].is_syntactically_equal(one(expr.sort)):
                return simp_st[1]
            if simp_st[1].is_syntactically_equal(one(expr.sort)):
                return simp_st[0]
            expr.subterms = simp_st
            return expr
        if expr.symbol.symbol == BuiltinFunctionSymbol.DIV:
            simp_st = (simplify(expr.subterms[0]), simplify(expr.subterms[1]))
            if simp_st[0].is_syntactically_equal(zero(expr.sort)):
                return zero(expr.sort)
            if simp_st[1].is_syntactically_equal(zero(expr.sort)):
                raise err.SemanticError("Division by Zero detected when \
            simplifying expression '{}' in division '{}'".format(expr.subterms[1], expr))
            if simp_st[1].is_syntactically_equal(one(expr.sort)):
                return simp_st[0]
            expr.subterms = simp_st
            return expr
        if expr.symbol.symbol == BuiltinFunctionSymbol.EXP:
            simp_st = simplify(expr.subterms[0])
            if simp_st.is_syntactically_equal(zero(expr.sort)):
                return one(expr.sort)
            expr.subterms = (simp_st,)
            return expr
    elif isinstance(expr, (Matrix, np.ndarray)):
        N, M = expr.shape
        for i in range(N):
            for j in range(M):
                expr[i, j] = simplify(expr[i, j])
        return expr
    elif isinstance(expr, AggregateCompoundTerm):
        expr.subterm = simplify(expr.subterm)
        return expr
    elif isinstance(expr, IfThenElse):
        expr.subterms = (simplify(expr.subterms[0]), simplify(expr.subterms[1]))
        return expr

    raise NotImplementedError("Can't handle expression {} yet".format(expr))



"""

"""
from ..syntax import Predicate, Function, Sort
from ..utils.algorithms import transitive_closure
from ..utils.hashing import consistent_hash
from .errors import ArityDLMismatch


class NullaryAtom:
    ARITY = 0

    def __init__(self, predicate):
        assert isinstance(predicate, Predicate)
        _check_arity("nullary atom", 0, predicate)
        self.name = predicate.symbol
        self.depth = 0
        self.hash = consistent_hash((self.__class__, self.name))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.name == other.name)

    def __repr__(self):
        return "{}".format(self.name)

    __str__ = __repr__

    def denotation(self, model):
        return model.primitive_denotation(self)


class GoalNullaryAtom(NullaryAtom):
    def __repr__(self):
        return "{}_g".format(self.name)

    __str__ = __repr__


class Concept:
    ARITY = 1

    def __init__(self, sort, size):
        assert isinstance(sort, str)
        self.sort = sort
        self.size = size

    def denotation(self, model):
        raise NotImplementedError()

    def flatten(self):
        raise NotImplementedError()


class Role:
    ARITY = 2

    def __init__(self, sort, size):
        assert len(sort) == self.ARITY
        self.sort = sort
        self.size = size

    def denotation(self, model):
        raise NotImplementedError()

    def flatten(self):
        raise NotImplementedError()


class UniversalConcept(Concept):
    def __init__(self, universal_sort):
        Concept.__init__(self, universal_sort, 0)
        self.hash = consistent_hash((self.__class__, ))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return self.__class__ is other.__class__

    def denotation(self, model):
        return model.compressed(model.primitive_denotation(self), self.ARITY)

    def __repr__(self):
        return '<universe>'

    __str__ = __repr__

    def flatten(self):
        return [self]


class EmptyConcept(Concept):
    def __init__(self, universal_sort):
        Concept.__init__(self, universal_sort, 0)
        self.hash = consistent_hash((self.__class__, ))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return self.__class__ is other.__class__

    def denotation(self, model):
        return model.compressed(model.primitive_denotation(self), self.ARITY)

    def __repr__(self):
        return '<empty>'

    __str__ = __repr__

    def flatten(self):
        return [self]


class NominalConcept(Concept):
    def __init__(self, name, sort):
        Concept.__init__(self, sort.name, 1)
        self.name = name
        self.hash = consistent_hash((self.__class__, self.name))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.name == other.name)

    def denotation(self, model):
        return model.compressed(model.primitive_denotation(self), self.ARITY)

    def __repr__(self):
        return "Nominal({})".format(self.name)

    __str__ = __repr__

    def flatten(self):
        return [self]


class PrimitiveConcept(Concept):
    def __init__(self, predicate):
        assert isinstance(predicate, (Predicate, Function, Sort))

        if isinstance(predicate, (Predicate, Function)):
            _check_arity("concept", 1, predicate)
            Concept.__init__(self, predicate.sort[0].name, 1)
            self.name = predicate.symbol  # This is a bit aggressive, but we assume that uniqueness of names

        else:
            Concept.__init__(self, predicate.name, 1)
            self.name = predicate.name
        self.hash = consistent_hash((self.__class__, self.name))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.name == other.name)

    def denotation(self, model):
        return model.compressed(model.primitive_denotation(self), self.ARITY)

    def __repr__(self):
        return "{}".format(self.name)

    __str__ = __repr__

    def flatten(self):
        return [self]


class GoalConcept(PrimitiveConcept):
    def __repr__(self):
        return "{}_g".format(self.name)

    __str__ = __repr__


class NotConcept(Concept):
    def __init__(self, c, universal_sort):
        assert isinstance(c, Concept)
        Concept.__init__(self, universal_sort.name, 1 + c.size)
        self.c = c
        self.hash = consistent_hash((self.__class__, self.c))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c == other.c)

    def denotation(self, model):
        return ~model.compressed_denotation(self.c)

    def __repr__(self):
        return 'Not({})'.format(self.c)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.c.flatten()


class AndConcept(Concept):
    def __init__(self, c1, c2, sort):
        assert isinstance(c1, Concept)
        assert isinstance(c2, Concept)
        Concept.__init__(self, sort, 1 + c1.size + c2.size)
        self.c1 = c1
        self.c2 = c2
        self.hash = consistent_hash((self.__class__, self.c1, self.c2))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c1 == other.c1 and
                self.c2 == other.c2)

    def denotation(self, model):
        ext_c1 = model.compressed_denotation(self.c1)
        ext_c2 = model.compressed_denotation(self.c2)
        return ext_c1 & ext_c2

    def __repr__(self):
        return 'And({},{})'.format(self.c1, self.c2)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.c1.flatten() + self.c2.flatten()


class OrConcept(Concept):
    def __init__(self, c1, c2, sort):
        assert isinstance(c1, Concept)
        assert isinstance(c2, Concept)
        Concept.__init__(self, sort, 1 + c1.size + c2.size)
        self.c1 = c1
        self.c2 = c2
        self.hash = consistent_hash((self.__class__, self.c1, self.c2))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c1 == other.c1 and
                self.c2 == other.c2)

    def denotation(self, model):
        ext_c1 = model.compressed_denotation(self.c1)
        ext_c2 = model.compressed_denotation(self.c2)
        return ext_c1 | ext_c2

    def __repr__(self):
        return 'Or({},{})'.format(self.c1, self.c2)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.c1.flatten() + self.c2.flatten()


class ExistsConcept(Concept):
    def __init__(self, r, c):
        assert isinstance(r, Role)
        assert isinstance(c, Concept)
        # The sort of an exists-concept is that of the first element of the relation
        Concept.__init__(self, r.sort[0], 1 + r.size + c.size)
        self.r = r
        self.c = c
        self.hash = consistent_hash((self.__class__, self.r, self.c))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c == other.c and
                self.r == other.r)

    def denotation(self, model):
        ext_c = model.uncompressed_denotation(self.c)
        ext_r = model.uncompressed_denotation(self.r)
        # result = [x for x in objects if [z for (y, z) in ext_r if y == x and z in ext_c]]
        result = set(x for x, y in ext_r if y in ext_c)
        return model.compressed(result, self.ARITY)

    def __repr__(self):
        return 'Exists({},{})'.format(self.r, self.c)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.r.flatten() + self.c.flatten()


class ForallConcept(Concept):
    def __init__(self, r, c):
        assert isinstance(r, Role)
        assert isinstance(c, Concept)
        # The sort of a forall-concept is that of the first element of the relation # TODO Check this
        Concept.__init__(self, r.sort[0], 1 + r.size + c.size)
        self.r = r
        self.c = c
        self.hash = consistent_hash((self.__class__, self.r, self.c))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c == other.c and
                self.r == other.r)

    def denotation(self, model):
        universe = model.universe()
        ext_c = model.uncompressed_denotation(self.c)
        ext_r = model.uncompressed_denotation(self.r)
        result = set()
        for x in universe:  # TODO COULD BE OPTIMIZED, E.G. IF R HAS EMPTY EXTENSION, ETC.
            ys = ext_c.union(y for y in universe if (x, y) not in ext_r)
            if len(universe) == len(ys):  # No need to compare the sets, as objects has max possible length
                result.add(x)
        return model.compressed(result, self.ARITY)

    def __repr__(self):
        return 'Forall({},{})'.format(self.r, self.c)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.r.flatten() + self.c.flatten()


class EqualConcept(Concept):
    def __init__(self, r1, r2, sort):
        assert isinstance(r1, Role)
        assert isinstance(r2, Role)
        Concept.__init__(self, sort, 1 + r1.size + r2.size)
        self.r1 = r1
        self.r2 = r2
        self.hash = consistent_hash((self.__class__, self.r1, self.r2))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.r1 == other.r1 and
                self.r2 == other.r2)

    def denotation(self, model):
        universe = model.universe()
        ext_r1 = model.uncompressed_denotation(self.r1)
        ext_r2 = model.uncompressed_denotation(self.r2)
        result = set()
        for x in universe:
            left = set(z for (y, z) in ext_r1 if y == x)
            right = set(z for (y, z) in ext_r2 if y == x)
            if left == right:
                result.add(x)
        return model.compressed(result, self.ARITY)

    def __repr__(self):
        return 'Equal({},{})'.format(self.r1, self.r2)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.r1.flatten() + self.r2.flatten()


class PrimitiveRole(Role):
    def __init__(self, predicate):
        assert isinstance(predicate, (Predicate, Function))
        _check_arity("role", 2, predicate)

        super().__init__([s.name for s in predicate.sort], 1)
        self.name = predicate.symbol

        # This is a bit aggressive, but we assume that predicate names are unique
        self.hash = consistent_hash((self.__class__, self.name))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.name == other.name)

    def denotation(self, model):
        return model.compressed(model.primitive_denotation(self), self.ARITY)

    def __repr__(self):
        return '{}'.format(self.name)

    __str__ = __repr__

    def flatten(self):
        return [self]


class GoalRole(PrimitiveRole):
    def __repr__(self):
        return "{}_g".format(self.name)

    __str__ = __repr__


class InverseRole(Role):
    def __init__(self, r):
        assert isinstance(r, Role)
        s1, s2 = r.sort
        super().__init__([s2, s1], 1 + r.size)
        self.r = r
        self.hash = consistent_hash((self.__class__, self.r))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.r == other.r)

    def denotation(self, model):
        ext_r = model.uncompressed_denotation(self.r)
        result = set((y, x) for (x, y) in ext_r)
        return model.compressed(result, self.ARITY)

    def __repr__(self):
        return 'Inverse({})'.format(self.r)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.r.flatten()


class StarRole(Role):
    def __init__(self, r):
        assert isinstance(r, Role)
        Role.__init__(self, r.sort, 1 + r.size)
        self.r = r
        self.hash = consistent_hash((self.__class__, self.r))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.r == other.r)

    def denotation(self, model):
        ext_r = model.uncompressed_denotation(self.r)
        result = set(transitive_closure(ext_r))
        return model.compressed(result, self.ARITY)

    def __repr__(self):
        return 'Star({})'.format(self.r)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.r.flatten()


class CompositionRole(Role):
    def __init__(self, r1, r2):
        assert isinstance(r1, Role)
        assert isinstance(r2, Role)
        Role.__init__(self, [r1.sort[0], r2.sort[1]], 1 + r1.size + r2.size)
        self.r1 = r1
        self.r2 = r2
        self.hash = consistent_hash((self.__class__, self.r1, self.r2))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.r1 == other.r1 and
                self.r2 == other.r2)

    def denotation(self, model):
        ext_r1 = model.uncompressed_denotation(self.r1)
        ext_r2 = model.uncompressed_denotation(self.r2)
        result = set()
        for a, b in ext_r1:
            for x, y in ext_r2:
                if b == x:
                    result.add((a, y))
                    break  # i.e. break the inner loop
        # for (x, u) in ext_r1:
        #     result.extend((x, z) for (y, z) in ext_r2 if u == y)
        return model.compressed(result, self.ARITY)

    def __repr__(self):
        return 'Composition({},{})'.format(self.r1, self.r2)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.r1.flatten() + self.r2.flatten()


class RestrictRole(Role):
    def __init__(self, r, c):
        assert isinstance(r, Role)
        assert isinstance(c, Concept)
        Role.__init__(self, r.sort, 1 + r.size + c.size)
        self.r = r
        self.c = c
        self.hash = consistent_hash((self.__class__, self.r, self.c))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return (hasattr(other, 'hash') and self.hash == other.hash and self.__class__ is other.__class__ and
                self.c == other.c and
                self.r == other.r)

    def denotation(self, model):
        ext_c = model.uncompressed_denotation(self.c)
        ext_r = model.uncompressed_denotation(self.r)
        result = set((x, y) for (x, y) in ext_r if y in ext_c)
        return model.compressed(result, self.ARITY)

    def __repr__(self):
        return 'Restrict({},{})'.format(self.r, self.c)

    __str__ = __repr__

    def flatten(self):
        return [self] + self.r.flatten() + self.c.flatten()


def _check_arity(term, expected_arity, predfun):
    if expected_arity != predfun.uniform_arity():
        raise ArityDLMismatch('Cannot create {} from predicate "{}"'.format(term, predfun))



from typing import Set
from .. import errors as err


class Sort:
    """ A logical sort (aka type)
        Sorts are uniquely identified by their name (i.e. we don't allow two sorts with different characteristics
        but the same name). Hence, implementation-wise, we can hash and compare them based on name alone.
    """
    def __init__(self, name, language, builtin=False):
        self._name = name
        self.language = language
        self._domain = set()
        self.builtin = builtin

    def __str__(self):
        return 'Sort({})'.format(self.name)

    __repr__ = __str__

    def __deepcopy__(self, memo):
        """ At the moment we forbid deep copies of this class, as they might be too expensive"""
        memo[id(self)] = self
        return self

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name and self.language == other.language

    @property
    def name(self):
        return self._name

    def contains(self, x):
        """ Return true iff the current sort contains a constant with the given value  """
        # TODO - Refactor this, we shouldn't be checking for two different ways of representing a value
        try:
            return x.symbol in self._domain
        except AttributeError:
            return x in self._domain

    def cast(self, x):
        # TODO - Refactor this, we shouldn't be checking for two different ways of representing a value
        try:
            if x.symbol in self._domain:
                return x.symbol
        except AttributeError:
            if x in self._domain:
                return x
            raise ValueError("Cast: Symbol '{}' does not belong to domain {}".format(x, self))
        return None

    def cardinality(self):
        return len(self._domain)

    def dump(self):
        return dict(name=self._name,
                    domain=list(self._domain))  # Copy the list

    def extend(self, constant):
        """ Extend the domain of the current sort, and recursively of the parent sorts, with a new constant. """
        self._domain.add(constant.symbol)
        for p in ancestors(self):
            p.extend(constant)

    def domain(self):
        return (self.language.get_constant(v) for v in self._domain)


class Interval(Sort):
    def __init__(self, name, lang, encode_fn, lower_bound=None, upper_bound=None):
        super().__init__(name, lang, builtin=True)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.encode = encode_fn

    def is_within_bounds(self, x):
        """ Check whether a given value is within the bounds of the interval """
        if self.lower_bound is None or self.upper_bound is None:
            raise err.SemanticError("Attempted to check for belonging to Interval type with no bounds set yet")
        return self.lower_bound <= x <= self.upper_bound

    def set_bounds(self, lower_bound, upper_bound):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def extend(self, constant):
        pass  # TODO ???

    def cardinality(self):
        return self.upper_bound - self.lower_bound + 1

    def cast(self, x):
        """ Casts the given value as an element of the current domain,
        or raise ValueError if it does not belong to it """
        # if isinstance(x, str):
        #     try:
        #         return getattr(self, x)  # TODO: WHAT IS THIS?? ANSWER: MADNESS
        #     except AttributeError:
        #         pass
        y = self.encode(x)  # can raise ValueError
        if not self.is_within_bounds(y):
            raise ValueError("Cast: Symbol '{}' (encoded '{}') outside of defined interval bounds".format(x, y))
        return y

    def contains(self, x):
        """ Returns true iff the given value belongs to the current domain """
        try:
            y = self.encode(x)
        except ValueError:
            return False
        return self.is_within_bounds(y)

    def _downcast(self, x):
        """ Check whether the given value belongs to the current sort _or_ can be downcasted to it.
        e.g. Integer.downcast(1.0) would return 1; whereas Integer.downcast(1.4) would return None. """
        # TODO (GFM) - Not sure we need this, and not sure whether the method works as it is
        # TODO (GFM) - If noone is using this we should remove it soon
        if self.contains(x):
            return self.encode(x)

        # Downcasting Python literals from their type to a subtype (i.e. Real to Integer) works
        # whenever the resulting instance of the subtype belongs
        # to the domain *and* Python equality over the subtype instance and the
        # supertype instance returns true.
        p = parent(self)
        while p is not None:
            try:
                z = p.cast(x)
            except ValueError:
                raise err.LanguageError()
            if z is not None and x != z:
                return None
            p = parent(p)

    def dump(self):
        return dict(name=self.name, domain=[self.lower_bound, self.upper_bound])


def inclusion_closure(s: Sort) -> Set[Sort]:
    """ Calculates the inclusion closure over given sort s """
    while s is not None:
        yield s
        s = parent(s)


def parent(s: Sort) -> Sort:
    """ Returns the direct parent of the given sort, or None if it is the root sort "object" """
    assert s in s.language.immediate_parent
    return s.language.immediate_parent[s]


def ancestors(s: Sort) -> Set[Sort]:
    """ Return all ancestor along the sort hierarchy of the given sort """
    assert s in s.language.ancestor_sorts
    return s.language.ancestor_sorts[s]


def children(s: Sort) -> Set[Sort]:
    """ Return the direct children of the given sort """
    result = set()
    for child, par in s.language.immediate_parent.items():
        if par is not None and par == s:
            result.add(child)
    return result


def int_encode_fn(x):
    if isinstance(x, float) and not x.is_integer():
        raise ValueError()  # We don't want 1.2 to get encoded as an int
    return int(x)


def float_encode_fn(x):
    return float(x)


def build_the_naturals(lang):
    the_nats = Interval('Natural', lang, int_encode_fn, 0, 2 ** 32 - 1)
    the_nats.builtin = True
    return the_nats


def build_the_integers(lang):
    the_ints = Interval('Integer', lang, int_encode_fn, -(2 ** 31 - 1), 2 ** 31 - 1)
    the_ints.builtin = True
    return the_ints


def build_the_reals(lang):
    reals = Interval('Real', lang, float_encode_fn, -3.40282e+38, 3.40282e+38)
    reals.builtin = True
    # the_reals.pi = scipy.constants.pi
    return reals




from ...fol import FirstOrderLanguage
from ...syntax.terms import Constant, Variable
from ..errors import UnableToGroundError


def enumerate_groundings(_: FirstOrderLanguage, symbols):
    """ """
    syms = []
    instantiations = []
    cardinality = 1
    for st in symbols:
        if isinstance(st, Constant):
            continue
        elif isinstance(st, Variable):
            if st.sort.builtin:
                raise UnableToGroundError(st,
                                          "Term is of built-in sort '{}', domain is too large!".format(st.sort.name))
            syms.append(st)
            instantiations.append(list(st.sort.domain()))
            cardinality *= len(instantiations[-1])
        else:
            raise UnableToGroundError(st, "Grounding of complex nested subterms is not implemented yet!")

    return cardinality, symbols, instantiations




from enum import Enum
from typing import List

from ..syntax import Tautology, CompoundTerm, Term, Constant, symref
from .errors import InvalidEffectError
from .. import theories as ths


class UniversalEffect:
    """ A forall-effect """

    def __init__(self, variables, effects, condition=Tautology()):
        self.variables = variables
        self.effects = effects
        self.condition = condition

    def __str__(self):
        return "forall ({}) : ({})".format(self.variables, ', '.join(self.effects))

    __repr__ = __str__


class SingleEffect:
    def __init__(self, condition):
        self.condition = condition

    def __str__(self):
        return "({} -> {})".format(self.condition, self.tostring())

    __repr__ = __str__

    def tostring(self):
        raise NotImplementedError("To be subclassed")


class AddEffect(SingleEffect):
    def __init__(self, atom, condition=Tautology()):
        super().__init__(condition)
        self.atom = atom

    def tostring(self):
        return "ADD({})".format(self.atom)


class DelEffect(SingleEffect):
    def __init__(self, atom, condition=Tautology()):
        super().__init__(condition)
        self.atom = atom

    def tostring(self):
        return "DEL({})".format(self.atom)


class LiteralEffect(SingleEffect):
    def __init__(self, lit, condition=Tautology()):
        super().__init__(condition)
        self.lit = lit

    def tostring(self):
        return "LIT({})".format(self.lit)


class FunctionalEffect(SingleEffect):
    def __init__(self, lhs, rhs, condition=Tautology()):
        super().__init__(condition)
        self.lhs = lhs
        self.rhs = rhs
        self.check_well_formed()

    def check_well_formed(self):
        if not isinstance(self.lhs, CompoundTerm):
            msg = "Error declaring FunctionalEffect: {}\n Invalid effect expression: \
            left hand side '{}' needs to be a functional term!".format(self.tostring(), self.lhs)
            raise InvalidEffectError(self, msg)

        if not isinstance(self.rhs, Term):
            msg = "Error declaring FunctionalEffect: {}\n Invalid effect expression: \
            right hand side '{}' needs to be a functional term!".format(self.tostring(), self.rhs)
            raise InvalidEffectError(self, msg)

    def tostring(self):
        return "{} := {}".format(self.lhs, self.rhs)


class IncreaseEffect(FunctionalEffect):
    def __init__(self, lhs, rhs, condition=Tautology()):
        super().__init__(lhs, rhs, condition)

        # MRJ: normalise rhs so it is easier to handle later on
        if type(self.rhs) == int:
            self.rhs = Constant(self.rhs, self.lhs.language.Integer)
        elif type(self.rhs) == float:
            self.rhs = Constant(self.rhs, self.lhs.language.Real)

    def check_well_formed(self):
        if not isinstance(self.lhs, CompoundTerm):
            msg = "Error declaring IncreaseEffect: {}\n Invalid effect expression: \
            left hand side '{}' needs to be a functional term!".format(self.tostring(), self.lhs)
            raise InvalidEffectError(self, msg)

        if not isinstance(self.rhs, Term) and type(self.rhs) not in [int, float]:
            msg = "Error declaring IncreaseEffect: {}\n Invalid increase expression: \
            right hand side '{}' needs to be a constant or functional term!".format(
                self.tostring(), self.lhs)
            raise InvalidEffectError(self, msg)


class OptimizationType(Enum):
    MINIMIZE = "minimize"
    MAXIMIZE = "maximize"

    def __str__(self):
        return self.value.lower()


class ProceduralEffect(SingleEffect):

    def __init__(self, input_: List[CompoundTerm], output: List[CompoundTerm]):
        super().__init__(Tautology())
        self.input = input_
        self.output = output

    def tostring(self):
        return "in: {}, out: {}".format(','.join([str(x) for x in self.input]),
                                        ','.join([str(x) for x in self.output]))


class ChoiceEffect(SingleEffect):

    def __init__(self, obj_type: OptimizationType, obj, variables: List[CompoundTerm], constraints=Tautology()):
        super().__init__(constraints)
        # MRJ: verify the effect is well formed
        self.obj = obj
        self.obj_type = obj_type
        self.variables = variables
        self.check_well_formed()

    def check_well_formed(self):
        if not isinstance(self.obj, CompoundTerm):
            msg = "Error declaring Choice Effect: {}\n Invalid objective expression: \
            expression to optimize needs to be a functional term!".format(self.tostring())
            raise InvalidEffectError(self, msg)

    def tostring(self):
        return "{} {}, vars: {} subject to: {}".format(self.obj_type,
                                                       self.obj, ','.join([str(x) for x in self.variables]),
                                                       self.condition)


class VectorisedEffect(SingleEffect):
    """ Action effects that modify the denotation of a vector (tuple) of terms """

    def __init__(self, lhs, rhs, condition=Tautology()):
        super().__init__(condition)
        self.lhs = lhs
        self.rhs = rhs
        self.check_well_formed()

    def tostring(self):
        return "VectorisedEffect({} := {})".format(self.lhs, self.rhs)

    def check_well_formed(self):
        if not hasattr(self.lhs, 'shape'):
            msg = "Error declaring VectorisedEffect, lhs needs needs to be\
                vector or matrix like"
            raise InvalidEffectError(self, msg)
        if self.lhs.shape[0] > 1:
            msg = "Error declaring VectorisedEffect, lhs needs needs to be\
                a vector in column order"
            raise InvalidEffectError(self, msg)
        if not hasattr(self.rhs, 'shape'):
            msg = "Error declaring VectorisedEffect, rhs needs needs to be\
                vector or matrix like"
            raise InvalidEffectError(self, msg)
        if self.lhs.shape == self.rhs.shape:
            msg = "Error declaring VectorisedEffect, lhs and rhs need to have\
                same dimensions"
            raise InvalidEffectError(self, msg)


class LinearEffect(SingleEffect):
    """
        Action effects that modify the denotation of a vector of (terms)  with
        the restriction of interactions between the expressions on the right hand
        side to be of the form:

            Ax + b
    """

    def __init__(self, y, a, x, b, condition=Tautology()):
        super().__init__(condition)
        self.y = y
        self.A = a
        self.x = x
        self.b = b
        self.check_well_formed()

    def tostring(self):
        return "LinearEffect({} := {} * {} + {})".format(self.y, self.A, self.x, self.b)

    def check_well_formed(self):
        if not hasattr(self.y, 'shape'):
            msg = "Error declaring VectorisedEffect, y needs needs to be\
                vector or matrix like"
            raise InvalidEffectError(self, msg)
        if self.y.shape[0] > 1:
            msg = "Error declaring VectorisedEffect, y needs needs to be\
                a vector in column order"
            raise InvalidEffectError(self, msg)
        if not hasattr(self.x, 'shape'):
            msg = "Error declaring VectorisedEffect, x needs needs to be\
                vector or matrix like"
            raise InvalidEffectError(self, msg)
        if self.y.shape == self.x.shape:
            msg = "Error declaring VectorisedEffect, y and x need to have\
                same dimensions"
            raise InvalidEffectError(self, msg)

        if not hasattr(self.A, 'shape'):
            msg = "Error declaring VectorisedEffect, coefficients matrix A needs to be\
                vector or matrix like"
            raise InvalidEffectError(self, msg)
        if self.y.shape[1] == self.A.shape[0]:
            msg = "Error declaring VectorisedEffect, y and x need to have\
                same dimensions"
            raise InvalidEffectError(self, msg)

        if not hasattr(self.b, 'shape'):
            msg = "Error declaring VectorisedEffect, coefficient vector b needs needs to be\
                vector or matrix like"
            raise InvalidEffectError(self, msg)
        if self.x.shape == self.b.shape:
            msg = "Error declaring VectorisedEffect, x and b need to have\
                same dimensions"
            raise InvalidEffectError(self, msg)


class BlackBoxEffect(SingleEffect):
    """
        Black box functional effect
    """

    def __init__(self, lhs, f, condition=Tautology()):
        super().__init__(condition)
        self.lhs = lhs
        self.function = f
        self.check_well_formed()

    def tostring(self):
        return "{} := {}".format(self.lhs, 'some function')

    def check_well_formed(self):
        """
            This method verifies
        """
        if not hasattr(self.lhs, 'shape'):
            msg = "Error declaring BlackBoxEffect, lhs needs needs to be\
                vector or matrix like"
            raise InvalidEffectError(self, msg)
        if self.lhs.shape[1] > 1:
            msg = "Error declaring BlackBoxEffect, lhs needs needs to be\
                a vector in column order"
            raise InvalidEffectError(self, msg)

        self.function.bind_to_language(self.lhs[0, 0].language)
        for k, y in enumerate(self.lhs[:, 0]):
            ys = symref(y)
            ok = symref(self.function.out_x[k].symbol)
            if ys != ok:
                msg = "Error declaring BlackBoxEffect, lhs {}-th symbol {} is not\
                matched by corresponding function output, {}".format(k, str(y), str(ok))
                raise InvalidEffectError(self, msg)


class OptimizationMetric:
    def __init__(self, opt_expression, opt_type):
        self.opt_expression = opt_expression
        self.opt_type = opt_type


def language(name="Unnamed FOL Language", theories=None):
    """ Create an FSTRIPS-oriented First-Order Language.
        This is a standard FOL with a few convenient add-ons.
    """
    # By default, when defining a FSTRIPS problem we use FOL with equality
    if theories is None:
        theories = [ths.Theory.EQUALITY]
    lang = ths.language(name, theories)
    lang.register_operator_handler("<<", Term, Term, FunctionalEffect)
    lang.register_operator_handler(">>", Term, Term, lambda lhs, rhs: FunctionalEffect(rhs, lhs))  # Inverted
    return lang



import itertools

import numpy as np

from .visitors import CollectFreeVariables, CollectVariables
from .terms import Term, Constant
from .formulas import CompoundFormula, Connective


def cast_to_closest_common_ancestor(lhs, rhs):
    if isinstance(lhs, Term):
        if isinstance(rhs, np.ndarray):
            # lhs is scalar, rhs is matrix
            return lhs.language.matrix([[lhs]], lhs.sort), rhs
        if not isinstance(rhs, Term):
            rhs = Constant(lhs.sort.cast(rhs), lhs.sort)
        return lhs, rhs
    if isinstance(lhs, np.ndarray):
        # lhs is matrix
        if isinstance(rhs, Term):
            return lhs, rhs.language.matrix([[rhs]])
    if not isinstance(rhs, Term):
        assert False  # this should not happen
    lhs = Constant(rhs.sort.cast(lhs), rhs.sort)

    return lhs, rhs


def infer_numeric_sort(value, language):
    # Note that this will only work in Python 3, which is fine.
    if isinstance(value, int):
        return language.Integer
    elif isinstance(value, float):
        return language.Real
    return None


def cast_to_number(rhs):
    assert isinstance(rhs, Constant)
    return rhs.symbol


def free_variables(formula):
    visitor = CollectFreeVariables()
    visitor.visit(formula)
    return list(visitor.free_variables)


def all_variables(formula):
    visitor = CollectVariables()
    visitor.visit(formula)
    return list(visitor.variables)


def flatten(formula):
    """ Flatten the given formula recursively, if possible, by applying associativity of and/or connectives
    Thus, a formula like (a or (b or (c or d))) becomes (a or b or c or d)
    """
    if not isinstance(formula, CompoundFormula) or formula.connective not in (Connective.And, Connective.Or):
        return formula
    return CompoundFormula(formula.connective, _flatten(formula, formula.connective))


def _flatten(formula, parent_connective):
    """ Return a list with all AST nodes of the given formula recursively flattened, i.e. where compound formulas
    with the given connective have been flattened themselves.
    """
    if not isinstance(formula, CompoundFormula) or formula.connective != parent_connective:
        return formula,  # (returns a tuple)
    return tuple(itertools.chain.from_iterable(_flatten(sub, parent_connective) for sub in formula.subformulas))



"""
    A module to create reachability-related logic programs from PDDL tasks
"""
import itertools

from ..syntax.builtins import symbol_complements
from ..syntax.ops import free_variables
from ..syntax import Formula, Atom, CompoundFormula, Connective, Term, Variable, Constant, Tautology, \
    BuiltinPredicateSymbol, QuantifiedFormula, symref, Quantifier
from ..syntax.sorts import parent
from ..fstrips import Problem, SingleEffect, UniversalEffect, AddEffect, DelEffect

SOLVABLE = "solvable"


def create_reachability_lp(problem: Problem):
    """ Return a reachability logic program, along with the symbol translation dictionary used to create it """
    lp = LogicProgram()
    compiler = ReachabilityLPCompiler(problem, lp)
    compiler.create()
    return lp, compiler.tr


class ReachabilityLPCompiler:
    def __init__(self, problem: Problem, lp):
        self.problem = problem
        self.lp = lp
        self.node_cache = dict()  # A cache of auxiliary subformulas
        self.aux_atom_count = 0
        self.tr = Translator()

    def gen_aux_atom(self, args=None):
        """ Return a new auxiliary atom with the given arguments """
        self.aux_atom_count += 1
        return self.lp_atom("__f{}".format(self.aux_atom_count), args)

    def create(self):
        problem, lang, lp = self.problem, self.problem.language, self.lp

        # Declare the PDDL objects with their types, e.g. with a fact "block(b1)".
        constants = lang.constants()
        _ = [lp.rule(self.lp_type_atom_from_term(c)) for c in constants]

        # Declare the type hierarchy, e.g. with a rule "object(X) :- block(X)."
        for s in lang.sorts:
            if not s.builtin:  # TODO Decide what to do with builtins
                p = parent(s)
                if p is not None:
                    lp.rule(self.lp_atom(p.name, [_var()]), [self.lp_atom(s.name, [_var()])])

        # Process all atoms in the initial state, e.g. "on(b1, b2)."
        for atom in problem.init.as_atoms():
            assert isinstance(atom, Atom)
            lp.rule(self.tarski_atom_to_lp_atom(atom))

        # Process all actions
        for _, action in problem.actions.items():
            # Construct the head and part of the body of the action atom, e.g. "move(X, Y) :- object(X), object(Y)"
            # Note that we need to capitalize the parameters of the action schema, as they are LP variables.
            action_atom = self.lp_atom(action.name, [make_variable_name(v.symbol) for v in action.parameters])
            body = [self.lp_type_atom_from_term(v) for v in action.parameters]

            # Add precondition atoms to the body
            body += self.process_formula(action.precondition)
            lp.rule(action_atom, body)

            # Now process the effects
            for eff in action.effects:
                head, body = self.process_effect(eff)
                if head is not None:
                    lp.rule(head, [action_atom] + body)

        # Process all derived predicates
        # TODO To be implemented yet
        assert not problem.derived_predicates

        # Process goal, e.g. "solvable :- on(a,b), on(b,c)." (note that the goal is always ground)
        body = self.process_formula(problem.goal)
        lp.rule(self.lp_atom(SOLVABLE), body)

    def process_formula(self, f: Formula):
        """ Process a given formula and return the corresponding LP rule body, along with declaring in the given LP
        any number of extra rules necessary to ensure equivalence of the body with the truth value of the formula.
        """
        ref = symref(f)
        res = self.node_cache.get(ref, None)
        # If the formula has been processed before and an auxiliary LP atom for it created, return that
        if res is not None:
            return res

        if isinstance(f, Tautology):
            return []

        elif isinstance(f, Atom):
            return [self.tarski_atom_to_lp_atom(f)]

        elif isinstance(f, CompoundFormula):
            if f.connective == Connective.And:
                return list(itertools.chain.from_iterable(self.process_formula(sub) for sub in f.subformulas))

            elif f.connective == Connective.Or:
                variables = free_variables(f)
                aux = self.gen_aux_atom([self.process_term(v) for v in variables])
                sub = [self.process_formula(s) for s in f.subformulas]
                for body in sub:
                    self.lp.rule(aux, body)
                return [aux]

            elif f.connective == Connective.Not:
                assert len(f.subformulas) == 1
                subf = f.subformulas[0]
                if not isinstance(subf, Atom):
                    raise RuntimeError("Negation of compound formulas not yet implemented")
                processed, = self.process_formula(subf)  # Unpack the length-1 list
                return [negate_lp_atom(processed)]

            else:
                raise RuntimeError('Unexpected connective "{}" within CompoundFormula "{}"'.format(f.connective, f))

        elif isinstance(f, QuantifiedFormula):
            if f.quantifier == Quantifier.Exists:
                aux = self.gen_aux_atom()  # Simply don't use any variable in the new atom
                self.lp.rule(aux, self.process_formula(f.formula))
                return [aux]

            else:
                assert f.quantifier == Quantifier.Forall
                raise RuntimeError('TODO')

        raise RuntimeError('Unexpected formula "{}" with type "{}"'.format(f, type(f)))

    def process_term(self, t: Term):
        """ Return the name of the LP constant corresponding to the given term. For instance, the variable "from" of
        type "place" will get transformed into a name "From" (capitalized), whereas the constant "b1" of type block will
        get transformed into a name "b1". We assume that the type information is used elsewhere. """
        if isinstance(t, Variable):
            return make_variable_name(t.symbol)  # lp_atom_from_term(t)
        elif isinstance(t, Constant):
            return t.symbol

        raise RuntimeError('Unexpected term "{}" with type "{}"'.format(t, type(t)))

    def process_effect(self, eff):
        """ Process a given effect and return the corresponding LP rule (a pair with head and body). For instance a
        conditional effect "p -> q(X)" will be transformed into a head q(X) and a body p.
        Additionally, declare in the given LP any number of extra rules necessary to ensure equivalence of the body
        with the truth value of the effect conditions (i.e. this will be mostly necessary for conditional effects).
        """
        assert isinstance(eff, (SingleEffect, UniversalEffect))
        if isinstance(eff, AddEffect):
            head = self.tarski_atom_to_lp_atom(eff.atom)
            return head, self.process_formula(eff.condition)
        elif isinstance(eff, DelEffect):
            return None, []  # Simply ignore the delete effects

        raise RuntimeError('Unexpected effect "{}" with type "{}"'.format(eff, type(eff)))

    def tarski_atom_to_lp_atom(self, atom: Atom):
        return self.lp_atom(atom.predicate.symbol, [self.process_term(sub) for sub in atom.subterms])

    def lp_type_atom_from_term(self, t: Term):
        """ Return a LP atom with type information from a given term, e.g. of the form block(b) for a constant,
        or block(X) for a variable. """
        if not isinstance(t, (Constant, Variable)):
            raise RuntimeError("Rechability logic program not ready for functional terms")
        name = make_variable_name(t.symbol) if isinstance(t, Variable) else _uncapitalize(t.symbol)
        return self.lp_atom(t.sort.name, [name])

    def lp_atom(self, symbol, args=None):
        args = args or []
        infix = False
        if isinstance(symbol, BuiltinPredicateSymbol):  # Special treatment for built-ins as =, !=, <, etc.
            symbol = symbol.value
            infix = True
        return LPAtom(self.tr.normalize(symbol), [self.tr.normalize(a) for a in args], infix=infix)


class LPAtom:
    def __init__(self, symbol: str, args=None, infix=False):
        self.symbol = symbol
        self.args = args or []
        self.infix = infix
        assert not infix or len(args) == 2

    def __str__(self):
        """ Return a string of the form 'symbol(arg1, ..., argn)', or 'symbol()', if args is empty """
        if self.infix:
            return "{} {} {}".format(self.args[0], self.symbol, self.args[1])
        arglist = ", ".join(str(arg) for arg in _ensure_list(self.args))
        return "{}({})".format(self.symbol, arglist)

    __repr__ = __str__


def negate_lp_atom(atom: LPAtom):
    negated = symbol_complements.get(atom.symbol, None)
    if negated is None:
        # We must have a non-builtin predicate symbol, hence we use the classical ASP negation symbol "-"
        assert not atom.infix
        negated = "-{}".format(negated)
    else:
        assert atom.infix
    return LPAtom(negated, atom.args, atom.infix)


def make_variable_name(name: str):
    if not name[0].isalpha():  # ASP variable names must start with a capital letter
        name = name[1:]
    return _capitalize(name)


def _capitalize(name: str):
    """ Capitalize the first letter of the name, e.g. for use as an ASP variable """
    return name.capitalize()


def _uncapitalize(name: str):
    """ Uncapitalize the first letter of the name, e.g. for use as an ASP constant """
    return name[0].lower() + name[1:]


def _ensure_list(elem):
    return elem if isinstance(elem, list) else [elem]


class Translator:
    def __init__(self):
        self.d = dict()
        self.inv = dict()

    def normalize(self, name: str):
        """ Translate a given name and keep the translation """
        translated = self.d.get(name, None)
        if translated is None:
            translated = sanitize(name)
            if name in self.inv or translated in self.inv:
                raise RuntimeError('Sanitization of STRIPS name for ASP purposes would create a name clash for key '
                                   '"{}"'.format(name))

            if translated != name:
                translated = self._insert(name, translated)
        return translated

    def _insert(self, k, v):
        self.d[k] = v
        self.inv[v] = k
        return v

    def back(self, name):
        """ Return the  unnormalized version of the given string, assuming it has been normalized on this translator """
        return self.inv.get(name, name)


class LogicProgram:
    def __init__(self):
        self.rules = []

    def rule(self, head, body=None):
        self.rules.append(_print_rule(head, body))

    def nrules(self):
        return len(self.rules)


class InFileLogicProgram:
    def __init__(self, fd):
        self.fd = fd

    def rule(self, head, body=None):
        self.fd.write(_print_rule(head, body) + "\n")


def _print_rule(head, body):
    assert body is None or isinstance(body, (list, tuple))
    return "{}.".format(head) if body is None else "{} :- {}.".format(head, _print_body(body))


def _print_body(body):
    return ", ".join(str(x) for x in body)


def sanitize(name: str):
    """ Normalize the given string by removing chars that are potentially problematic for the LP solver """
    return name.replace("-", "__")


def _var(i=0, lowercase=False):
    """ Return a distinct variable name for each given value of i """
    alphabet = "XYZABCDEFGHIJKLMNOPQRSTUFW"
    res = alphabet[i] if i < len(alphabet) else "X{}".format(i)
    return res.lower() if lowercase else res




from ...syntax import Atom, CompoundFormula, Connective, Constant, CompoundTerm
from ...syntax.builtins import BuiltinPredicateSymbol

from .errors import TransformationError


def unpack_element_subterms(element):
    elems = []
    for c in element.subterms:
        if not isinstance(c, Constant):
            raise TransformationError("transform-to-ground-atoms", element, "Cannot unpack non-ground atom")
        elems.append(c.symbol)
    return elems


def unpack_atom(atom):
    assert isinstance(atom, Atom)

    if atom.predicate.builtin:
        if not atom.predicate.symbol == BuiltinPredicateSymbol.EQ:
            raise TransformationError("transform-to-ground-atoms", atom, "Cannot unpack non-equality atom")
        x = [x for x in atom.subterms if isinstance(x, CompoundTerm)]
        c = [c for c in atom.subterms if isinstance(c, Constant)]
        if len(x) != 1 or len(c) != 1:
            raise TransformationError("transform-to-ground-atoms", atom,
                                      "Can only use functional atoms of the form f(c) = d, where c and d are constants")
        x, c = x[0], c[0]
        return tuple([x.symbol.symbol] + unpack_element_subterms(x) + [c.symbol])

    else:
        return tuple([atom.predicate.symbol] + unpack_element_subterms(atom))


def transform_to_ground_atoms(phi):
    """ Take a formula expected to be a conjunction of ground atoms and return the list of ground atoms,
    each of them as a tuple of the form (on, a, b).
    Raise an exception if the formula is not a conjunction of ground atoms. """
    if not isinstance(phi, Atom) and not (isinstance(phi, CompoundFormula) and phi.connective == Connective.And):
        raise TransformationError("transform-to-ground-atoms", phi,
                                  "Only conjunctions of ground atoms can be transformed")

    if isinstance(phi, Atom):  # Base case of the recursion
        return [unpack_atom(phi)]

    atoms = []
    for sub in phi.subformulas:
        atoms += transform_to_ground_atoms(sub)
    return atoms




class Reaction:
    """ A (possibly lifted) reaction """

    def __init__(self, language, name, parameters, condition, effect):
        self.name = name
        self.language = language
        self.parameters = parameters
        self.condition = condition
        self.effect = effect
        self._check_well_formed()

    def _check_well_formed(self):
        pass

    def ident(self):
        params = ', '.join([str(o) for o in self.parameters])
        return '{}({})'.format(self.name, params)

    def dump(self):
        return dict(name=self.name,
                    params=[par.dump() for par in self.parameters],
                    condition=self.condition.dump(),
                    effect=[eff.dump() for eff in self.effect.dump()])

    def __str__(self):
        tokens = ['reaction {}:'.format(self.name),
                  'cond: ({})'.format(self.condition),
                  'eff: ({})'.format(self.effect)]
        return '\n'.join(tokens)




import itertools

from ...fstrips.contingent import Sensor
from ...syntax import create_substitution, TermSubstitution
from ...util import IndexDictionary
from .elements import process_expression
from . import instantiation


class SensorGrounder:

    def __init__(self, prob, index):
        self.problem = prob
        self.L = self.problem.language
        self.index = index
        self.problem.ground_sensors = IndexDictionary()
        self.schemas = list(self.problem.sensors.values())
        self.sensors_generated = 0

    def __str__(self):
        return 'Sensors generated: {}'.format(self.sensors_generated)

    def calculate_sensors(self):
        # @TODO: this is pretty much the same code as that of grounding actions
        # this NEEDS to be refactored, so we ground action elements in a more
        # moudular and component oriented fashion
        for act_schema in self.schemas:
            K, syms, substs = instantiation.enumerate_groundings(self.L, act_schema.parameters)
            for values in itertools.product(*substs):
                subst = create_substitution(syms, values)

                op = TermSubstitution(self.L, subst)
                g_prec = process_expression(self.L, act_schema.condition, op)
                g_obs = process_expression(self.L, act_schema.obs, op)
                self.problem.ground_sensors.add(Sensor(self.L, act_schema.name, [], g_prec, g_obs))
            self.sensors_generated += K



"""
    Visitors implementing diverse aspects of FSTRIPS problems translation,
    analysis and compilation.
"""
from ..syntax.temporal import ltl
from ..syntax.formulas import CompoundFormula, Atom, QuantifiedFormula
from ..syntax.terms import CompoundTerm
from ..syntax import symref


class FluentHeuristic:
    action_effects = 1
    precondition = 2
    constraint = 3


class FluentSymbolCollector:
    """
        This visitor collects CompoundTerms which are candidates to become
        state variables.
    """

    def __init__(self, lang, fluents, statics, mode: FluentHeuristic):
        self.mode = mode
        self.lang = lang
        self.fluents = fluents
        self.statics = statics
        self.under_next = False
        self.visited = set()

    def reset(self):
        self.visited = set()

    def _visit_action_effect_formula(self, phi):

        if isinstance(phi, CompoundFormula):
            _ = [self.visit(f) for f in phi.subformulas]

        elif isinstance(phi, QuantifiedFormula):
            self.visit(phi.formula)

        elif isinstance(phi, Atom):
            if not phi.predicate.builtin:
                self.fluents.add(symref(phi))
            else:
                _ = [self.visit(f) for f in phi.subterms]

        elif isinstance(phi, CompoundTerm):
            # print("Compound Term: {}, {}, {}".format(str(phi), phi.symbol, phi.symbol.builtin))
            if not phi.symbol.builtin:
                self.fluents.add(symref(phi))
            else:
                _ = [self.visit(f) for f in phi.subterms]

    def _visit_constraint_formula(self, phi):
        if isinstance(phi, ltl.TemporalCompoundFormula) and phi.connective == ltl.TemporalConnective.X:
            old_value = self.under_next
            self.under_next = True
            _ = [self.visit(f) for f in phi.subformulas]
            self.under_next = old_value

        elif isinstance(phi, CompoundFormula):
            old_visited = self.visited.copy()
            _ = [self.visit(f) for f in phi.subformulas]
            delta = self.visited - old_visited
            # print('Fluents: {}'.format([str(x) for x in self.fluents]))
            # print('Delta: {}'.format([str(x) for x in delta]))
            if any(f in self.fluents for f in delta):
                # print("Fluency propagates")
                for f in delta:
                    self.fluents.add(f)

        elif isinstance(phi, QuantifiedFormula):
            self.visit(phi.formula)

        elif isinstance(phi, Atom):
            if not phi.predicate.builtin:
                self.visited.add(symref(phi))
            if self.under_next:
                if not phi.predicate.builtin:
                    self.fluents.add(symref(phi))
            else:
                self.statics.add(symref(phi))
            _ = [self.visit(f) for f in phi.subterms]

        elif isinstance(phi, CompoundTerm):
            if not phi.symbol.builtin:
                self.visited.add(symref(phi))
            if self.under_next:
                if not phi.symbol.builtin:
                    self.fluents.add(symref(phi))
            else:
                self.statics.add(symref(phi))
            _ = [self.visit(f) for f in phi.subterms]

    def _visit_precondition_formula(self, phi):
        if isinstance(phi, CompoundFormula):
            _ = [self.visit(f) for f in phi.subformulas]

        elif isinstance(phi, QuantifiedFormula):
            self.visit(phi.formula)

        elif isinstance(phi, Atom):
            self.statics.add(symref(phi))

        elif isinstance(phi, CompoundTerm):
            self.statics.add(symref(phi))

    def visit(self, phi):
        """
            Visitor method to sort atoms and terms into the
            "fluent" and "static" categories. Note that a given
            symbol can be in both sets, this means that it gets
            "votes" as static and fluent... the post_process() method
            is meant to settle the issue (and potentially allow for
            more ellaborate/clever heuristics).

            NB: at the moment we're trawling all (possibly lifted)
            sub-expressions, this is intentional.
        """
        if self.mode == FluentHeuristic.action_effects:
            self._visit_action_effect_formula(phi)
        elif self.mode == FluentHeuristic.constraint:
            self._visit_constraint_formula(phi)
        else:
            assert self.mode == FluentHeuristic.precondition
            self._visit_precondition_formula(phi)



"""
    NNF Rewriter
"""
import copy

from ..formulas import neg, QuantifiedFormula, Quantifier, CompoundFormula, Connective


class NNFTransformation:
    """
        This class rewrites the input formula phi into an equivalent formula
        in NNF
    """

    def __init__(self, phi, do_copy=True):
        self.blueprint = None
        if do_copy:
            self.blueprint = copy.deepcopy(phi)
        else:
            self.blueprint = phi
        self.nnf = None

    def _convert(self, phi):
        if isinstance(phi, CompoundFormula):
            if phi.connective == Connective.Not:
                p = phi.subformulas[0]
                if isinstance(p, QuantifiedFormula):
                    if p.quantifier == Quantifier.Exists:
                        p.quantifier = Quantifier.Forall
                    else:
                        assert p.quantifier == Quantifier.Forall
                        p.quantifier = Quantifier.Exists
                    p.formula = self._convert(neg(p.formula))
                    return p
                elif isinstance(p, CompoundFormula):  # De Morgan
                    if p.connective == Connective.Not:
                        return p.subformulas[0]  # eliminate \neg \neg
                    elif p.connective == Connective.And:
                        p.connective = Connective.Or
                    else:
                        assert p.connective == Connective.Or
                        p.connective = Connective.And

                    new_sub = [self._convert(neg(p.subformulas[0])),
                               self._convert(neg(p.subformulas[1]))]
                    p.subformulas = tuple(new_sub)
                    return p
                else:
                    return phi  # nothing to do
            else:
                assert phi.connective == Connective.And or phi.connective == Connective.Or
                new_sub = [self._convert(phi.subformulas[0]),
                           self._convert(phi.subformulas[1])]
                phi.subformulas = tuple(new_sub)
                return phi
        elif isinstance(phi, QuantifiedFormula):
            phi.formula = self._convert(phi.formula)
            return phi
        else:
            return phi

    def convert(self):
        self.nnf = self._convert(self.blueprint)

    @staticmethod
    def rewrite(phi, do_copy=True):
        trans = NNFTransformation(phi, do_copy)
        trans.convert()
        return trans




from ... errors import DuplicateDefinition, UndefinedElement, SyntacticError


class DuplicateReactionDefinition(DuplicateDefinition):
    pass


class UndefinedReaction(UndefinedElement):
    pass


class DuplicateDifferentialConstraintDefinition(DuplicateDefinition):
    pass


class UndefinedDifferentialConstraint(UndefinedElement):
    pass


class InvalidDifferentialConstraintDefinition(SyntacticError):

    def __init__(self, culprit, msg):
        error_msg = "Invalid Differential Constraint\n'{}' {}".format(str(culprit), msg)
        super().__init__(self, msg=error_msg)




from enum import Enum
from ... import errors as err
from ..formulas import Formula, Connective, CompoundFormula, lor


class TemporalConnective(Enum):
    X, F, G, U, R = range(5)

    def __str__(self):
        return self.name


class TemporalCompoundFormula(CompoundFormula):

    def __init(self, conn, sub):
        super().__init__(conn, sub)

    def _check_well_formed(self):
        if any(not isinstance(f, Formula) for f in self.subformulas):
            raise err.LanguageError("Wrong argument types for compound formula: '{}' ".format(self.subformulas))

        if self.connective == Connective.Not or \
                self.connective in (TemporalConnective.X, TemporalConnective.F, TemporalConnective.G):
            if len(self.subformulas) != 1:
                raise err.LanguageError("{} admits only one subformula".format(str(self.connective)))
        elif len(self.subformulas) < 2:
            raise err.LanguageError("{} requires at least two subformulas".format(str(self.connective)))

    def __str__(self):
        if self.connective == Connective.Not or \
                self.connective in (TemporalConnective.X, TemporalConnective.F, TemporalConnective.G):
            assert len(self.subformulas) == 1
            return "{} ({})".format(self.connective, str(self.subformulas[0]))

        inner = " {} ".format(self.connective).join(str(f) for f in self.subformulas)
        return "({})".format(inner)


def X(arg):
    return TemporalCompoundFormula(TemporalConnective.X, [arg])


def F(arg):
    return TemporalCompoundFormula(TemporalConnective.F, [arg])


def G(arg):
    return TemporalCompoundFormula(TemporalConnective.G, [arg])


def U(*args):
    return TemporalCompoundFormula(TemporalConnective.U, args)


def R(*args):
    return TemporalCompoundFormula(TemporalConnective.R, args)


def W(*args):
    return lor(U(args), G(args[0]))
    # return CompoundFormula(TemporalConnective.W, args)



import logging
from collections import defaultdict

from .common import load_tpl
from ..model import ExtensionalFunctionDefinition
from ..syntax import Tautology, Contradiction, Atom, CompoundTerm, CompoundFormula, QuantifiedFormula, \
    Term, Variable, Constant, Formula
from ..syntax.sorts import parent, Interval, ancestors

from ._fstrips.common import tarski_to_pddl_type, get_requirements_string
from ..fstrips import create_fstrips_problem, language, FunctionalEffect, AddEffect, DelEffect, IncreaseEffect,\
    UniversalEffect

from ._fstrips.reader import FStripsParser

# Leave the next import so that it can be imported from the outside without warnings of importing a private module
# pylint: disable=unused-import
from ._fstrips.reader import ParsingError


class FstripsReader:

    def __init__(self, raise_on_error=False, theories=None):
        self.problem = create_fstrips_problem(language=language(theories=theories))
        self.parser = FStripsParser(self.problem, raise_on_error)

    def read_problem(self, domain, instance):
        self.parse_domain(domain)
        self.parse_instance(instance)
        return self.problem

    def parse_file(self, filename, start_rule):
        logging.info('Parsing filename "{}" from grammar rule "{}"'.format(filename, start_rule))
        domain_parse_tree, _ = self.parser.parse_file(filename, start_rule)
        self.parser.visit(domain_parse_tree)

    def parse_domain(self, filename):
        self.parse_file(filename, 'domain')

    def parse_instance(self, filename):
        self.parse_file(filename, 'problem')

    def parse_string(self, string, start_rule):
        logging.info('Parsing custom string from grammar rule "{}"'.format(start_rule))
        parse_tree, _ = self.parser.parse_string(string, start_rule)
        logging.info("Processing AST")
        return self.parser.visit(parse_tree)


_TAB = " " * 4

action_tpl = """
    (:action {name}
     :parameters ({parameters})
     :precondition {precondition}
     :effect {effect}
    )
"""

derived_tpl = """
    (:derived ({name} {parameters})
              {formula}
    )
"""


def print_objects(constants):
    """ Print a PDDL object declaration with the given objects.
    Objects are sorted by name and grouped by type, and types sorted by name as well """
    constants_by_sort = defaultdict(list)
    for c in constants:
        constants_by_sort[c.sort.name].append(c.symbol)

    elements = []
    for sort in sorted(constants_by_sort.keys()):
        sobjects = " ".join(sorted(constants_by_sort[sort]))
        elements.append("{} - {}".format(sobjects, sort))

    return linebreaks(elements, indentation=2, indent_first=False)


def linebreaks(elements, indentation, indent_first):
    def indent_while_iterating():
        for i, element in enumerate(elements, 0):
            idt = 0 if i == 0 and not indent_first else indentation
            yield indent(element, idt)

    return "\n".join(indent_while_iterating())


def print_init(problem):
    elements = []
    # e.g. (= (value c0) 0)
    for signature, definition in problem.init.function_extensions.items():
        if not isinstance(definition, ExtensionalFunctionDefinition):
            continue  # Ignore intensionally defined symbols
        fname = signature[0]
        for point, value in definition.data.items():
            elements.append("(= ({} {}) {})".format(fname, print_term_ref_list(point), value))

    # e.g. (clear b1)
    for signature, definition in problem.init.predicate_extensions.items():
        assert isinstance(definition, set)
        predname = signature[0]
        for point in definition:
            elements.append("({} {})".format(predname, print_term_ref_list(point)))

    return linebreaks(elements, indentation=2, indent_first=False)


def print_goal(problem):
    return print_formula(problem.goal, 1)


def print_domain_bounds(problem):
    lang = problem.language
    bounds = []
    for sort in lang.sorts:
        if not sort.builtin and isinstance(sort, Interval):
            assert lang.has_sort('Integer')
            if lang.Integer in ancestors(sort):
                bounds.append("({} - int[{}..{}])".format(sort.name, sort.lower_bound, sort.upper_bound))
            elif lang.Real in ancestors(sort):
                pass  # TODO

    if not bounds:  # No bounded type was found
        return ""

    inner = "\n".join(indent(b, 2) for b in bounds)
    return "(:bounds\n{})".format(inner)


def print_problem_constraints(problem):
    # pylint: disable=unused-argument
    return ""  # TODO


def print_metric(metric):
    return '(:metric {type} {exp})'.format(type=metric.opt_type.value,
                                           exp=print_term(metric.opt_expression))


def print_problem_metric(problem):
    return print_metric(problem.plan_metric) if problem.plan_metric else ''


class FstripsWriter:

    def __init__(self, problem):
        self.problem = problem
        self.lang = problem.language

    def write(self, domain_filename, instance_filename, domain_constants=None):
        domain_constants = domain_constants or []
        self.write_domain(domain_filename, domain_constants)
        self.write_instance(instance_filename, domain_constants)

    def print_domain(self, constant_objects=None):
        tpl = load_tpl("fstrips_domain.tpl")
        content = tpl.format(
            header_info="",
            domain_name=self.problem.domain_name,
            requirements=" ".join(get_requirements_string(self.problem)),
            types=self.get_types(),
            functions=self.get_functions(),
            predicates=self.get_predicates(),
            actions=self.get_actions(),
            derived=self.get_derived_predicates(),
            constants=print_objects(constant_objects if constant_objects else set()),
        )
        return content

    def write_domain(self, filename, constant_objects):
        with open(filename, 'w') as file:
            file.write(self.print_domain(constant_objects))

    def print_instance(self, constant_objects=None):
        tpl = load_tpl("fstrips_instance.tpl")

        # Only objects which are not declared in the domain file need to be printed in the instance file
        constant_objs_set = set(constant_objects) if constant_objects else set()
        instance_objects = [c for c in self.problem.language.constants() if c not in constant_objs_set]

        content = tpl.format(
            header_info="",
            domain_name=self.problem.domain_name,
            problem_name=self.problem.name,

            objects=print_objects(instance_objects),
            init=print_init(self.problem),
            goal=print_goal(self.problem),
            constraints=print_problem_constraints(self.problem),
            domain_bounds=print_domain_bounds(self.problem),
            metric=print_problem_metric(self.problem),
        )
        return content

    def write_instance(self, filename, constant_objects):
        with open(filename, 'w') as file:
            file.write(self.print_instance(constant_objects))

    def get_types(self):
        res = []
        for t in self.lang.sorts:
            if t.builtin or t == self.lang.Object:
                continue  # Don't declare builtin elements
            tname = tarski_to_pddl_type(t)
            p = parent(t)
            if p:
                res.append("{} - {}".format(tname, tarski_to_pddl_type(p)))
            else:
                res.append(tname)
        return ("\n" + _TAB * 2).join(res)

    def get_functions(self):
        res = []
        for fun in self.lang.functions:
            if fun.builtin:
                continue  # Don't declare builtin elements
            domain_str = build_signature_string(fun.domain)
            codomain_str = tarski_to_pddl_type(fun.codomain)
            res.append("({} {}) - {}".format(fun.symbol, domain_str, codomain_str))
        return ("\n" + _TAB * 2).join(res)

    def get_predicates(self):
        res = []
        for fun in self.lang.predicates:
            if fun.builtin:
                continue  # Don't declare builtin elements
            domain_str = build_signature_string(fun.sort)
            res.append("({} {})".format(fun.symbol, domain_str))
        return ("\n" + _TAB * 2).join(res)

    def get_actions(self):
        return "\n".join(self.get_action(a) for a in self.problem.actions.values())

    @staticmethod
    def get_action(a):
        base_indentation = 1
        return action_tpl.format(
            name=a.name,
            parameters=print_variable_list(a.parameters),
            precondition=print_formula(a.precondition, base_indentation),
            effect=print_effects(a.effects, base_indentation)
        )

    def get_derived_predicates(self):
        return "\n".join(self.get_derived(d) for d in self.problem.derived_predicates.values())

    @staticmethod
    def get_derived(d):
        return derived_tpl.format(
            name=d.predicate.symbol,
            parameters=print_variable_list(d.parameters),
            formula=print_formula(d.formula))


def build_signature_string(domain):
    if not domain:
        return ""

    return " ".join("?x{} - {}".format(i, tarski_to_pddl_type(t)) for i, t in enumerate(domain, 1))


def print_variable_list(parameters):
    return " ".join("?{} - {}".format(p.symbol, p.sort.name) for p in parameters)


def print_formula(formula, indentation=0):
    # pylint: disable=unused-argument
    assert isinstance(formula, Formula)
    if isinstance(formula, Tautology):
        return "(and )"
    elif isinstance(formula, Contradiction):
        return "(= 0 1)"  # PDDL HACK =)
    elif isinstance(formula, Atom):
        return print_atom(formula)
    elif isinstance(formula, CompoundFormula):
        return "({} {})".format(formula.connective, print_formula_list(formula.subformulas))

    elif isinstance(formula, QuantifiedFormula):
        vars_ = print_variable_list(formula.variables)
        # e.g. (exists (?x - object) (and (= ?x 2)))
        return '({} ({}) {})'.format(formula.quantifier, vars_, print_formula(formula.formula))
    raise RuntimeError("Unexpected element type: {}".format(formula))


def print_effects(effects, indentation=0):
    if not effects:
        return "(and )"
    return "(and\n{})".format("\n".join(print_effect(e, indentation + 1) for e in effects))


def print_unconditional_effect(eff, indentation=0):
    functional = isinstance(eff, FunctionalEffect)
    increase = isinstance(eff, IncreaseEffect)

    if increase:
        return indent("(increase {} {})".format(print_term(eff.lhs), print_term(eff.rhs)), indentation)
    elif functional:
        return indent("(assign {} {})".format(print_term(eff.lhs), print_term(eff.rhs)), indentation)
    elif isinstance(eff, AddEffect):
        return indent("{}".format(print_atom(eff.atom)), indentation)
    elif isinstance(eff, DelEffect):
        return indent("(not {})".format(print_atom(eff.atom)), indentation)
    elif isinstance(eff, UniversalEffect):
        effect_str = (print_effect(eff.effects[0]) if len(eff.effects) == 1 else print_effects(eff.effects))
        return indent("(forall ({}) {})".format(print_variable_list(eff.variables), effect_str),
                      indentation)

    raise RuntimeError("Unexpected element type: {}".format(eff))


def print_effect(eff, indentation=0):
    conditional = not isinstance(eff.condition, Tautology)

    if conditional:
        return indent(
            "(when {} {})".format(print_formula(eff.condition), print_unconditional_effect(eff)),
            indentation)
    else:
        return print_unconditional_effect(eff, indentation)


def print_term(term):
    assert isinstance(term, Term)
    if isinstance(term, Variable):
        return "?{}".format(term.symbol)
    elif isinstance(term, CompoundTerm):
        return "({} {})".format(term.symbol.symbol, print_term_list(term.subterms))
    elif isinstance(term, Constant):
        return "{}".format(term.symbol)
    raise RuntimeError("Unexpected element type: {}".format(term))


def print_atom(atom):
    assert isinstance(atom, Atom)
    return "({} {})".format(atom.predicate.symbol, print_term_list(atom.subterms))


def print_term_list(terms):
    return " ".join(print_term(t) for t in terms)


def print_term_ref_list(termrefs):
    return " ".join(print_term(t.expr) for t in termrefs)


def print_formula_list(formulas):
    return " ".join(print_formula(f) for f in formulas)


def indent(text, indentation):
    return (indentation * _TAB) + text



import logging
import os
import subprocess


def count_file_lines(filename):  # Might be a bit faster with a call to "wc -l"
    """ Count the number of lines in a given file """
    i = 0
    with open(filename) as f:
        for i, _ in enumerate(f, 1):
            pass
    return i


def remove_duplicate_lines(filename):
    """ Removes in-place any duplicate line in the file. Will also reorder the lines as a side-effect """
    subprocess.call(['sort', '-u', '-o', filename, filename])


def read_file(filename):
    """ Read a file, line by line, ignoring end-of-line characters"""
    with open(filename) as f:
        for line in f:
            yield line.rstrip('\n')


def execute(command, **kwargs):
    stdout, stderr = kwargs.get("stdout", None), kwargs.get("stderr", None)
    if isinstance(stdout, str):
        stdout = open(stdout, 'w')

    if isinstance(stderr, str):
        stderr = open(stderr, 'w')

    cwd = kwargs["cwd"] if "cwd" in kwargs else os.getcwd()

    msg = 'Executing "{}" on directory "{}"'.format(' '.join(command), cwd)
    if stdout:
        msg += '. Standard output redirected to "{}"'.format(stdout.name)
    if stderr:
        msg += '. Standard error redirected to "{}"'.format(stderr.name)
    logging.info(msg)

    retcode = subprocess.call(command, cwd=cwd, stdout=stdout, stderr=stderr)

    if stdout:
        stdout.close()

    if stderr:
        stderr.close()

    if stderr is not None and os.path.getsize(stderr.name) == 0:  # Delete error log if empty
        os.remove(stderr.name)

    return retcode




from collections import OrderedDict

from ..problem import Problem
from .reaction import Reaction
from .differential_constraints import DifferentialConstraint
from . import errors as err
from .. import fstrips as fs


class HybridProblem(Problem):
    """ A Functional STRIPS Hybrid problem """

    def __init__(self):
        super().__init__()
        self.reactions = OrderedDict()
        self.differential_constraints = OrderedDict()

    def reaction(self, name, parameters, condition, effect):
        if name in self.reactions:
            raise err.DuplicateReactionDefinition(name, self.reactions[name])
        self.reactions[name] = Reaction(self.language, name, parameters, condition, effect)
        return self.reactions[name]

    def has_reaction(self, name):
        return name in self.reactions

    def get_reaction(self, name):
        if not self.has_reaction(name):
            raise err.UndefinedReaction(name)
        return self.reactions[name]

    def differential_constraint(self, name, parameters, condition, variate, ode):
        if name in self.reactions:
            raise err.DuplicateDifferentialConstraintDefinition(name, self.reactions[name])
        self.differential_constraints[name] = DifferentialConstraint(self.language, name, parameters, condition,
                                                                     variate, ode)
        return self.differential_constraints[name]

    def has_differential_constraint(self, name):
        return name in self.differential_constraints

    def get_differential_constraint(self, name):
        if not self.has_differential_constraint(name):
            raise err.UndefinedDifferentialConstraint(name)
        return self.differential_constraints[name]

    def get_symbols(self, pv, ev, cv):
        super().get_symbols(pv, ev, cv)
        for _, react in self.reactions.items():
            pv.visit(react.condition)
            eff = react.effect
            if isinstance(eff, (fs.AddEffect, fs.DelEffect)):
                ev.visit(eff.atom)
            elif isinstance(eff, (fs.FunctionalEffect, fs.ChoiceEffect)):
                ev.visit(eff.lhs)
            elif isinstance(eff, fs.BlackBoxEffect):
                for yk in eff.lhs[:, 0]:
                    ev.visit(yk)
            else:
                raise RuntimeError("Effect type '{}' cannot be analysed".format(type(eff)))
        for _, dc in self.differential_constraints.items():
            pv.visit(dc.condition)
            pv.visit(dc.variate)

    def __str__(self):
        return 'FSTRIPS Hybrid Problem "{}", domain "{}"'.format(self.name, self.domain_name)

    __repr__ = __str__



from ...errors import TarskiError


class SubstitutionError(TarskiError):
    def __init__(self, phi, subst, msg=None):
        msg = msg or 'unspecified error'
        super().__init__('Error applying substitution "{}" to formula "{}": {}'.format(str(subst), str(phi), msg))


class TransformationError(TarskiError):
    def __init__(self, name, phi, msg=None):
        msg = msg or 'unspecified error'
        super().__init__('Error applying transformation "{}" to formula "{}": {}'.format(name, str(phi), msg))




from collections import OrderedDict
from enum import Enum
from typing import List

from .. import errors as err
from .terms import Variable, Term
from .util import termlists_are_equal, termlist_hash
from .predicate import Predicate


class Connective(Enum):
    """" A logical connective """
    And, Or, Not = range(3)

    def __str__(self):
        return self.name.lower()


class Quantifier(Enum):
    """" A logical quantifier """
    Exists, Forall = range(2)
    to_string = ["exists", "forall"]

    def __str__(self):
        return self.name.lower()


class Formula:
    """ A first-order logical formula. """
    def __str__(self):
        raise NotImplementedError()  # To be subclassed

    def __and__(self, rhs):
        return land(self, rhs)

    def __or__(self, rhs):
        return lor(self, rhs)

    def __invert__(self):
        return neg(self)

    def __gt__(self, rhs):
        return implies(self, rhs)

    def __eq__(self, other):
        raise NotImplementedError()  # To be subclassed

    def __hash__(self):
        raise NotImplementedError()  # To be subclassed

    def hash(self):
        # Define a shortcut for uniformity with the Term class
        return self.__hash__()

    def is_syntactically_equal(self, other):
        """ Return true if this formula and other are strictly syntactically equivalent.
        This is equivalent to self == other, but is provided for reasons of uniformity with the Term class.
        """
        return self.__eq__(other)


class Tautology(Formula):
    def __str__(self):
        return "T"

    def __eq__(self, other):
        return self.__class__ is other.__class__

    def __hash__(self):
        return hash(self.__class__)


class Contradiction(Formula):
    def __str__(self):
        return "F"

    def __eq__(self, other):
        return self.__class__ is other.__class__

    def __hash__(self):
        return hash(self.__class__)


class CompoundFormula(Formula):
    """ A set of formulas combined through some logical connective """

    def __init__(self, connective, subformulas):
        super().__init__()
        self.connective = connective
        self.subformulas = subformulas
        self._check_well_formed()

    def _check_well_formed(self):
        if any(not isinstance(f, Formula) for f in self.subformulas):
            raise err.LanguageError("Wrong argument types for compound formula: '{}' ".format(self.subformulas))

        if self.connective == Connective.Not:
            if len(self.subformulas) != 1:
                raise err.LanguageError("Negation admits only one subformula")
        elif len(self.subformulas) < 2:
            raise err.LanguageError("Conjunction and disjunction require at least two subformulas")

    def __str__(self):
        if self.connective == Connective.Not:
            assert len(self.subformulas) == 1
            return "{} ({})".format(self.connective, str(self.subformulas[0]))

        inner = " {} ".format(self.connective).join(str(f) for f in self.subformulas)
        return "({})".format(inner)

    def __eq__(self, other):
        return self.__class__ is other.__class__ and \
               self.connective == other.connective and \
               self.subformulas == other.subformulas

    def __hash__(self):
        return hash((self.__class__, self.connective, self.subformulas))


class QuantifiedFormula(Formula):
    def __init__(self, quantifier: Quantifier, variables: List[Variable], formula: Formula):
        self.quantifier = quantifier
        self.variables = variables
        self.formula = formula
        self._check_well_formed()

    def _check_well_formed(self):
        if len(self.variables) == 0:
            raise err.LanguageError("Quantified formula with no variable")

    def __str__(self):
        vars_ = ', '.join(str(x) for x in self.variables)
        return '{} {} : ({})'.format(self.quantifier, vars_, self.formula)

    def __eq__(self, other):
        return self.__class__ is other.__class__ \
               and self.quantifier == other.quantifier \
               and termlists_are_equal(self.variables, other.variables) \
               and self.formula == other.formula

    def __hash__(self):
        return hash((self.__class__, self.quantifier, termlist_hash(self.variables), self.formula))


top = Tautology()
bot = Contradiction()


def _to_binary_tree(args, connective):
    assert len(args) > 1
    phi = CompoundFormula(connective, (args[-2], args[-1]))
    for arg in reversed(args[:-2]):
        phi = CompoundFormula(connective, (arg, phi))
    return phi


def _create_compound(args, connective, flat):
    if flat:
        return CompoundFormula(connective, args)
    return _to_binary_tree(args, connective)


def land(*args, flat=False):
    """ Create an and-formula with the given subformulas. If binary is true, the and-formula will be shaped as a binary
     tree (e.g. (...((p1 and p2) and p3) and ...))), otherwise it will have a flat structure. This is an implementation
     detail, but might be relevant performance-wise when dealing with large structures """
    return _create_compound(args, Connective.And, flat)


def lor(*args, flat=False):
    """ Create an or-formula with the given subformulas. If binary is true, the or-formula will be shaped as a binary
    tree (e.g. (...((p1 or p2) or p3) or ...))), otherwise it will have a flat structure. This is an implementation
    detail, but might be relevant performance-wise when dealing with large structures """
    return _create_compound(args, Connective.Or, flat)


def neg(phi):
    return CompoundFormula(Connective.Not, [phi])


def implies(phi, psi):
    return lor(neg(phi), psi)


def equiv(phi, psi):
    # MRJ: I choose the form below in the code over
    #
    # lor(land(phi,psi), land(neg(psi),neg(phi)))
    #
    # as I find it easier to transform into a
    # given normal form.
    return land(implies(phi, psi), implies(psi, phi))


def forall(*args):
    return _quantified(Quantifier.Forall, *args)


def exists(*args):
    return _quantified(Quantifier.Exists, *args)


def _quantified(quantifier, *args):
    """ Create a quantified formula.

        'args' is expected to be of the form [v1, ..., vn, f], where v_i are the variables
        and f is the quantified formula.
    """
    if len(args) < 2:
        raise err.LanguageError('Quantified formula needs at least two arguments')

    variables, formula = args[:-1], args[-1]

    if not isinstance(formula, Formula):
        raise err.LanguageError('Illformed arguments for quantified formula: {}'.format(args))

    if not all(isinstance(x, Variable) for x in variables):
        raise err.LanguageError('Illformed arguments for quantified formula: {}'.format(args))

    return QuantifiedFormula(quantifier, variables, args[-1])


class Atom(Formula):
    """ A first-order atom. """

    def __init__(self, predicate, arguments):
        super().__init__()
        self.predicate = predicate
        self.subterms = arguments
        self._check_well_formed()

    def _check_well_formed(self):
        head = self.predicate

        if not isinstance(head, Predicate):
            raise err.LanguageError("Incorrect atom head: '{}' ".format(head))

        # Check arities match
        if len(self.subterms) != self.predicate.arity:
            raise err.ArityMismatch(head, self.subterms)

        language = head.language

        # Check arguments are all terms of the appropriate type and matching language
        for arg, expected_sort in zip(self.subterms, head.sort):
            if not isinstance(arg, Term):
                raise err.LanguageError("Wrong argument for atomic formula: '{}' ".format(arg))

            if arg.language != language:
                raise err.LanguageMismatch(arg, arg.language, language)

            if not language.is_subtype(arg.sort, expected_sort):
                raise err.SortMismatch(arg, arg.sort, expected_sort)

    def __str__(self):
        return '{}({})'.format(self.predicate.symbol, ','.join([str(t) for t in self.subterms]))
    __repr__ = __str__

    def __eq__(self, other):
        return self.__class__ is other.__class__ and \
               self.predicate == other.predicate and \
               termlists_are_equal(self.subterms, other.subterms)

    def __hash__(self):
        return hash((self.__class__, self.predicate, termlist_hash(self.subterms)))


class VariableBinding:
    """ A VariableBinding contains a set of logical variables which are _bound_ in some formula or term """
    def __init__(self, variables=None):
        variables = variables or []
        # An (ordered) map between variable name and the variable itself:
        self.variables = OrderedDict((v.symbol, v) for v in variables)
        self._idx = 0
        self._v_values = list(self.variables.values())
        self.index_ = {v.symbol: i for i, v in enumerate(variables)}

    def __len__(self):
        return len(self.variables)

    def __getitem__(self, index):
        return self._v_values[index]

    def add(self, variable: Variable):
        other = self.variables.get(variable.symbol, None)
        if other is not None:
            raise err.DuplicateVariableDefinition(variable, other)
        self.variables[variable.symbol] = variable
        self._v_values += [variable]
        self.index_[variable.symbol] = len(self.index_)

    def get(self, name):
        var = self.variables.get(name, None)
        if var is None:
            raise err.UndefinedVariable(name)
        return var

    def index(self, name):
        idx = self.index_.get(name, None)
        if idx is None:
            raise err.UndefinedVariable(name)
        return idx

    def merge(self, binding):
        """ Merge the given binding into the current binding, inplace """
        raise NotImplementedError()

    @staticmethod
    def empty():
        return VariableBinding()

    def vars(self):
        return list(self.variables.values())

    def __iter__(self):
        self._v_values = [v for _, v in self.variables.items()]
        self._idx = 0
        return self

    def __next__(self):
        if self._idx == len(self._v_values):
            raise StopIteration()
        self._idx += 1
        return self._v_values[self._idx - 1]



import logging
import shutil
import tempfile
from collections import defaultdict

from ..errors import CommandNotFoundError, ExternalCommandError
from ..utils import command as cmd


def run_clingo(lp):
    gringo = shutil.which("gringo")
    if gringo is None:
        raise CommandNotFoundError("gringo")

    with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as f:
        _ = [print(str(r), file=f) for r in lp.rules]
        filename = f.name

    logging.debug('Using gringo binary found in "{}"'.format(gringo))
    with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as f:
        # Option "-t" enforces an easier-to-parse textual output. Warnings could also be supressed with
        # option "-Wno-atom-undefined"
        retcode = cmd.execute([gringo, "-t", filename], stdout=f)
        model_filename = f.name
        if retcode != 0:
            raise ExternalCommandError("Gringo exited with code {}".format(retcode))

    return model_filename


def parse_model(filename, symbol_mapping):
    tr = symbol_mapping
    model = defaultdict(set)
    with open(filename, "r") as f:
        for line in f:
            data = line.rstrip(' \n.').rstrip(')')
            components = data.split('(')
            if len(components) == 1:
                symbol = tr.back(components[0])
                model[symbol].add(())
            elif len(components) == 2:
                symbol, arguments = components
                model[tr.back(symbol)].add(tuple(tr.back(s) for s in arguments.split(',')))
            else:
                # No nested terms expected, so there should be at most 2 components
                raise RuntimeError('Unexpected line "{}" in Clingo solution file'.format(line))

    return model



import logging
from collections import deque

from .model import ForwardSearchModel


class BreadthFirstSearch:
    """ Apply Breadth-First search to a FSTRIPS problem """
    def __init__(self, model: ForwardSearchModel, max_expansions=-1):
        self.model = model
        self.max_expansions = max_expansions

    def run(self):
        return self.search(self.model.init())

    def search(self, s0):
        # create obj to track state space
        space = SearchSpace()

        iteration = 0
        num_goals_found = 0

        open_ = deque()  # fifo-queue storing the nodes which are next to explore
        closed = set()

        open_.append(make_root_node(s0))

        node_id = 0
        while open_:
            iteration += 1
            # logging.debug("brfs: Iteration {}, #unexplored={}".format(iteration, len(open_)))

            node = open_.popleft()
            is_goal = self.model.is_goal(node.state)

            space.expand(node)

            # we manage the closed list here to allow the parents update
            if node.state in closed:
                continue

            closed.add(node.state)

            node_id += 1

            # exploring the node or if it is a goal node extracting the plan
            if is_goal:
                num_goals_found += 1
                logging.info("Goal found after {} expansions. Number of goal states found: {}".format(
                    node_id, num_goals_found))

            if 0 <= self.max_expansions <= node_id:
                logging.info("Max. expansions reached. # expanded: {}, # goals: {}".format(node_id, num_goals_found))
                return space

            for operator, successor_state in self.model.successors(node.state):
                open_.append(make_child_node(node, operator, successor_state))

        logging.info("Search space exhausted. # expanded: {}, # goals: {}".format(node_id, num_goals_found))
        space.complete = True
        return space


class SearchNode:
    def __init__(self, state, parent, action):
        self.state = state
        self.parent = parent
        self.action = action


class SearchSpace:
    """ A representation of a search space / transition system corresponding to some planning problem """
    def __init__(self):
        self.nodes = set()
        self.last_node_id = 0
        self.complete = False  # Whether the state space contains all states reachable from the initial state

    def expand(self, node: SearchNode):
        self.nodes.add(node)


def make_root_node(state):
    """ Construct the initial root node without parent nor action """
    return SearchNode(state, None, None)


def make_child_node(parent_node, action, state):
    """ Construct an child search node """
    return SearchNode(state, parent_node, action)




import numpy as np

from ..builtins import BuiltinFunctionSymbol as BFS


def normal(mu, sigma):
    try:
        normal_func = mu.language.get_function(BFS.NORMAL)
    except AttributeError:
        try:
            normal_func = mu.language.get_function(BFS.NORMAL)
        except AttributeError:
            return np.random.normal(mu, sigma)
    return normal_func(mu, sigma)


def gamma(shape, scale):
    try:
        gamma_func = shape.language.get_function(BFS.GAMMA)
    except AttributeError:
        try:
            gamma_func = scale.language.get_function(BFS.GAMMA)
        except AttributeError:
            return np.random.gamma(shape, scale)
    return gamma_func(shape, scale)




from ..syntax import Predicate, Function, Constant, termlists_are_equal, termlist_hash
from ..fstrips import AddEffect, DelEffect, FunctionalEffect


class StateVariableLite:
    """ A state variable is nothing else than a CompoundTerm or Atom which is expected to change its
    value, along with a particular instantiation of its subterms.
    Note: This is a lightweight version of the StateVariable class above, hoping that it can eventually replace it.
    """

    def __init__(self, symbol, binding):
        assert isinstance(symbol, (Predicate, Function))
        assert all(isinstance(c, Constant) for c in binding)
        self.symbol = symbol
        self.binding = binding

    def __hash__(self):
        return hash((self.symbol, termlist_hash(self.binding)))

    def __eq__(self, other):
        return self.symbol == other.symbol and termlists_are_equal(self.binding, other.binding)

    def __str__(self):
        return '{}({})'.format(self.symbol.symbol, ','.join(map(str, self.binding)))

    __repr__ = __str__


def approximate_symbol_fluency(problem):
    """ Sort out all predicate and function symbols of a given problem into static and fluent symbols.

    Fluent symbols are those whose denotation might (in principle) be affected by the effect of some action schema;
    the rest are static symbols. This methods works on a purely syntactic basis, and does not take into account any
    reachability information, nor whether the action schemas can be detected as unapplicable at preprocessing. This
    for instance means that an action a(x) with precondition False and atomic add effect "p(x)" will be taken as
    evidence that predicate symbol "p" is fluent. It also means that "fluency" is only considered at the symbol
    level (i.e. "p" is either fluent or static), not at the state variable level (where e.g. p(a) could be seen as
    static, and p(b) as fluent).

    The results of this method overapproaximate the set of symbol fluents (i.e. false positives are possible, false
    negatives are not).

    This can be used as a basis for a finer-grained analysis that can be performed e.g. with the ASP-based reachability
    grounder.
    """
    lang = problem.language
    all_symbols = set(lang.functions) | set(lang.predicates)

    # All symbols appearing on some action, process or event effect are fluent
    fluents = set(_compute_effect_head_symbol(eff) for action in problem.actions.values() for eff in action.effects)

    # The rest of symbols are considered static
    statics = set(s for s in all_symbols if not s.builtin and s not in fluents)
    return fluents, statics


def _compute_effect_head_symbol(effect):
    """ Return the symbol affected by a given effect. """
    if isinstance(effect, (AddEffect, DelEffect)):
        return effect.atom.predicate
    elif isinstance(effect, FunctionalEffect):
        return effect.lhs.symbol
    else:
        raise RuntimeError('Unknown effect type "{}"'.format(type(effect)))




import logging
from .version import __version__, __version_info__
from .fol import FirstOrderLanguage
from .theories import language
from .syntax import Function, Predicate, Formula, Term, Constant, Variable
from .errors import LanguageError
from . import fstrips
logging.getLogger(__name__).addHandler(logging.NullHandler())




from ...errors import TarskiError, DuplicateDefinition, UndefinedElement


class ObservationExpressivenessMismatch(TarskiError):
    def __init__(self, sensor, msg=None):
        msg = msg or 'mismatch between the expressiveness of the specified sensor and that supported by the language!'
        super().__init__(
            'Sensing action: {}: {}\nObservation should be literal (Atom, or CompoundFormula) and it is: {}'.format(
                sensor.name, msg, type(sensor.obs)))


class DuplicateSensorDefinition(DuplicateDefinition):
    pass


class UndefinedSensor(UndefinedElement):
    pass



"""
    Rewrite formulas into prenex normal form
"""
from ..symrefs import symref
from ..formulas import CompoundFormula, QuantifiedFormula, Connective, Quantifier, lor
from ..transform.nnf import NNFTransformation
from ..transform import term_substitution

from .errors import TransformationError


class PrenexTransformation:
    """
        This class rewrites the input formula phi into an equivalent formula
        in Prenex NNF

    """

    def __init__(self, lang, phi, do_copy=True):
        self.L = lang
        self.blueprint = NNFTransformation.rewrite(phi, do_copy).nnf
        self.prenex = None

    def _merge_quantified_subformulas(self, lhs, rhs, renaming=True):
        assert isinstance(lhs, QuantifiedFormula)
        assert isinstance(rhs, QuantifiedFormula)
        new_variables = {(x.symbol, x.sort.name): x for x in lhs.variables}
        subst = {}
        for y in rhs.variables:
            key_y = (y.symbol, y.sort.name)
            if key_y not in new_variables:
                new_variables[key_y] = y
            else:
                if renaming:
                    y2 = self.L.variable("{}'".format(y.symbol), y.sort)
                    subst[y] = y2
                    new_variables[(y2.symbol, y2.sort.name)] = y2
        if len(subst) > 0:
            rhs.formula = term_substitution(self.L, rhs.formula, subst, inplace=True)
        new_phi = QuantifiedFormula(lhs.quantifier, list(new_variables.values()), lor(lhs.formula, rhs.formula))
        return new_phi

    def _nest_quantifiers(self, out_q, out_vars, out_phi, inner_q, inner_vars, conn, lhs, rhs):
        """
            Note that out_phi is either lhs or rhs, we have the parameter duplicated so we
            can preserve the ordering of subformulas
        """
        in_vars_dict = {(x.symbol, x.sort.name): x for x in inner_vars}
        new_out_vars = []
        subst = {}
        for y in out_vars:
            key_y = (y.symbol, y.sort.name)
            if key_y not in in_vars_dict:
                new_out_vars.append(y)
            else:
                y2 = self.L.variable("{}'".format(y.symbol), y.sort)
                subst[symref(y)] = y2
                new_out_vars.append(y2)
        if len(subst) > 0:
            term_substitution(self.L, out_phi, subst, inplace=True)
        phi = CompoundFormula(conn, tuple([lhs, rhs]))
        inner = QuantifiedFormula(inner_q, inner_vars, phi)
        return QuantifiedFormula(out_q, new_out_vars, inner)

    def _convert(self, phi):
        if isinstance(phi, CompoundFormula):
            return self.convert_compound(phi)

        elif isinstance(phi, QuantifiedFormula):
            return self.convert_quantified(phi)

        else:
            return phi

    def convert_compound(self, phi):
        assert isinstance(phi, CompoundFormula)
        if phi.connective == Connective.Not:
            if isinstance(phi.subformulas[0], QuantifiedFormula):
                raise TransformationError('prenex', phi, 'Subformula is not in NNF!')

            if isinstance(phi.subformulas[0], CompoundFormula):
                raise TransformationError('prenex', phi, 'Subformula is not in NNF!')

            return phi
        else:
            # and/or
            return self.convert_and_or(phi)

    def convert_and_or(self, phi):
        lhs, rhs = self._convert(phi.subformulas[0]), self._convert(phi.subformulas[1])
        is_quant_lhs = isinstance(lhs, QuantifiedFormula)
        is_quant_rhs = isinstance(rhs, QuantifiedFormula)
        if not is_quant_lhs and not is_quant_rhs:
            return phi
        elif is_quant_lhs and is_quant_rhs:
            # both parts of the formula are quantified
            if lhs.quantifier == rhs.quantifier:
                return self._merge_quantified_subformulas(lhs, rhs)
                # GFM: Code below commented out, as it is unreachable!
                # if lhs.quantifier == Quantifier.Exist:
                #     if phi.connective == Connective.And:
                #         return self._merge_quantified_subformulas(lhs, rhs)
                #     assert phi.connective == Connective.Or
                #     return self._merge_quantified_subformulas(lhs, rhs, False)  # no renaming
                # assert lhs.quantifier == Quantifier.Forall
                # if phi.connective == Connective.Or:
                #     return self._merge_quantified_subformuls(lhs, rhs)
                # return self._merge_quantified_subformulas(lhs, rhs, False)
            # we have different quantifiers, we apply the null quantifier
            # equivalence

            if rhs.quantifier == Quantifier.Exists and lhs.quantifier == Quantifier.Forall:
                return self._nest_quantifiers(Quantifier.Exists, rhs.variables, rhs.formula,
                                              Quantifier.Forall, lhs.variables,
                                              phi.connective, lhs.formula, rhs.formula)
            return self._nest_quantifiers(Quantifier.Exists, lhs.variables, lhs.formula,
                                          Quantifier.Forall, rhs.variables,
                                          phi.connective, lhs.formula, rhs.formula)
        # \forall ( P \lor Q(x)) \equiv P \lor \forall x Q(x)
        # \exists (P \land Q(x)) \equiv P \land \exists x Q(x)
        elif is_quant_rhs:
            return _merge_mixed_subformulas(rhs.quantifier, rhs.variables, lhs, phi.connective, rhs.formula)
        else:  # is_quant_lhs
            assert is_quant_lhs
            return _merge_mixed_subformulas(lhs.quantifier, lhs.variables, lhs.formula, phi.connective, rhs)

    def convert_quantified(self, phi):
        assert isinstance(phi, QuantifiedFormula)
        phi.formula = self._convert(phi.formula)
        if isinstance(phi.formula, QuantifiedFormula):
            if phi.formula.quantifier == phi.quantifier:  # absorb
                new_variables = [x for x in phi.variables]
                for x in phi.formula.variables:
                    new_variables.append(x)
                phi.formula.variables = tuple(new_variables)
                return phi.formula
            if phi.formula.quantifier == Quantifier.Exists:  # push up existential
                phi.quantifier, phi.formula.quantifier = phi.formula.quantifier, phi.quantifier
                phi.variables, phi.formula.variables = phi.formula.variables, phi.variables
                phi.formula = self._convert(
                    phi.formula)  # reordering the quantifiers may trigger further quantifier reordering
            return phi
        else:
            return phi

    def convert(self):
        self.prenex = self._convert(self.blueprint)

    @staticmethod
    def rewrite(lang, phi, do_copy=True):
        trans = PrenexTransformation(lang, phi, do_copy)
        trans.convert()
        return trans


def _merge_mixed_subformulas(quant, variables, lhs, conn, rhs):
    new_phi = QuantifiedFormula(quant, variables, CompoundFormula(conn, tuple([lhs, rhs])))
    return new_phi




class Action:
    """ A (possibly lifted) planning action """

    def __init__(self, language, name, parameters, precondition, effects):
        self.name = name
        self.language = language
        self.parameters = parameters
        self.precondition = precondition
        self.effects = effects

    def __lt__(self, other):
        return self.name < other.name

    def dump(self):
        return dict(name=self.name,
                    params=[par.dump() for par in self.parameters],
                    precondition=self.precondition.dump(),
                    effects=[eff.dump() for eff in self.effects.dump()])

    def ident(self):
        paramlist = "{}".format(','.join("{}: {}".format(p.symbol, p.sort.name) for p in self.parameters))
        return '{}({})'.format(self.name, paramlist)

    def __str__(self):
        tokens = ['{}:'.format(self.ident()),
                  'pre=({})'.format(self.precondition),
                  'eff=({})'.format(' & '.join(str(eff) for eff in self.effects))]
        return '\n'.join(tokens)
